//**********************************************************************
// This source code is provided as an example.  No warranty is expressed
// or implied.  It is your responsibility to determine its suitability
// to your application, to modify it as necessary, and to validate
// its correctness.
//
// Licensed Materials - Property of IBM
// (c) Copyright IBM Corporation 2000. All Rights Reserved.
//
// US Government Users Restricted Rights - Use, duplication, or
// disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
//**********************************************************************
// Revision / History
// Defect     Date       Who      Description
// ---------- 01/03/2000 ldbader  Module created
//**********************************************************************

package com.ibm.etill.ldbcardcassette;

import  java.text.NumberFormat;
import  java.sql.PreparedStatement;
import  java.sql.ResultSet;
import  java.sql.SQLException;
import  java.util.Enumeration;
import  java.util.Hashtable;
import  java.util.Locale;
import  java.util.Vector;


import  com.ibm.etill.framework.admin.Admin;
import  com.ibm.etill.framework.admin.AccountAdmin;
import  com.ibm.etill.framework.archive.Archivable;
import  com.ibm.etill.framework.archive.ETillArchive;
import  com.ibm.etill.framework.cassette.AdminRequest;
import  com.ibm.etill.framework.cassette.APIResponse;
import  com.ibm.etill.framework.log.ErrorLog;
import  com.ibm.etill.framework.log.ETillCassetteException;
import  com.ibm.etill.framework.log.Trace;
import  com.ibm.etill.framework.payapi.Batch;
import  com.ibm.etill.framework.payapi.Credit;
import  com.ibm.etill.framework.payapi.ETillAbortOperation;
import  com.ibm.etill.framework.payapi.Order;
import  com.ibm.etill.framework.payapi.ParameterTable;
import  com.ibm.etill.framework.payapi.Payment;
import  com.ibm.etill.framework.payapi.Transaction;
import  com.ibm.etill.framework.supervisor.Supervisor;


import  netverify.remote.client.Connection;
import  netverify.remote.client.DriverManager;
import  netverify.remote.client.TransactionStatement;
import  netverify.remote.util.ICVException;

/***********************************************************************
* The Account object is responsible for managing the relationship
* between the merchant and the payment engine.  It uses the
* cassette's account configuration table to create account objects and
* verifies that the configuration is syntactically and semantically
* correct.  It is responsible for making connections to the payment
* engine and exchanging data between the cassette and the payment
* engine.
***********************************************************************/
public final class LdbCardAccount
   implements LdbCardConstants, Archivable {

   private static final String COPYRIGHT =
      "(c) Copyright IBM Corporation 2000 - All Rights Reserved.";

   //<gutter>
   //</gutter>

   /********************************************************************
   * Constants for core engine parameter keys.
   * <p>
   * These keys are used when passing key/value pairs to the
   * payment engine.
   ********************************************************************/
   private static final String ICV_KEY_ACCOUNT   = "Account";
   private static final String ICV_KEY_ADDRESS   = "Address";
   private static final String ICV_KEY_AMOUNT    = "Amount";
   private static final String ICV_KEY_APPROVAL  = "ApprovalCode";
   private static final String ICV_KEY_CLERK     = "Clerk";
   private static final String ICV_KEY_COMMENT   = "Comment";
   private static final String ICV_KEY_EXPIRY    = "Expiry";
   private static final String ICV_KEY_REASON    = "DeclineReason";
   private static final String ICV_KEY_STATUS    = "TransactionStatus";
   private static final String ICV_KEY_TYPE      = "TransactionType";
   private static final String ICV_KEY_ZIP_CODE  = "ZipCode";

   /********************************************************************
   * Constants for core engine status values.
   * <p>
   * These values are returned when when you get the status from
   * a transaction statement that has been executed.
   ********************************************************************/
   private static final String ICV_VAL_APPROVED  = "Approved";

   /********************************************************************
   * Constants for core engine parameter values.
   * <p>
   * These values are define for the TransactionType key.
   ********************************************************************/
   private static final String ICV_VAL_BOOK      = "Book";
   private static final String ICV_VAL_CREDIT    = "Credit";
   private static final String ICV_VAL_SETTLE    = "Settle";
   private static final String ICV_VAL_SHIP      = "Ship";
   private static final String ICV_VAL_VOID      = "Void";
   private static final String ICV_VAL_VOID_C    = "CreditVoid";

   /********************************************************************
   * Constants to wrap the framework order numbers for the
   * core engine.
   * <p>
   * By using the following tags, the framework order number
   * is recorded in the database.  The value has the form
   * <CMO>ordernumber:paymentnumber</CMO> for payments or
   * <CMO>ordernumber:credit</CMO> for credits.
   ********************************************************************/
   private static final String ICV_ORDER_PREFIX      = "<CMO>";
   private static final String ICV_ORDER_SEPARATOR   = ":";
   private static final String ICV_ORDER_SUFFIX      = "</CMO>";

   /********************************************************************
   * This Hashtable contains the entire list of all the merchant
   * accounts.  In this implementation, by definition, there is one
   * and only one account for each merchant and only one currency
   * for each account.  Furthermore, there is one and only one
   * open batch for each account.
   ********************************************************************/
   private static Hashtable    clsAccountList    = new Hashtable();

   /********************************************************************
   * This method is called by the cassette when the framework asks
   * the cassette to read its configuration.  It is a static method
   * that initalizes all the existing account objects for the cassette.
   * <p>
   * For this cassette, the method reads every row from the account
   * configuration table and constructs an account object for each row.
   * The account objects are kept in a Hashtable (a static variable)
   * that can be used to retrieve the appropriate account object
   * when required.
   * <p>
   * Since, by definition, there can only be one account per merchant,
   * the merchant number is used as a key in the Hashtable and if
   * the database table contains more than one account for any given
   * merchant, an exception is thrown.
   *
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public static void resurrectAccounts (
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.resurrectAccounts()"
         );
      }
      // Create SQL statment to read the account configuration table.
      String varSelect =
         " SELECT * FROM " +
         ETillArchive.geteTillDatabaseOwner() + "." + TBL_ACCOUNT;

      // Obtain an ETillArchive object for reading the database.
      ETillArchive varDatabase = new ETillArchive();

      // Lock the database to prevent JDBC races.
      synchronized (ETillArchive.getArchiveLock()) {
         try {

            // Trace database read.
            if (Trace.isAnyoneTracing()) {
               Trace.traceDatabaseRead(
                  TRACE_ID,
                  "\n" + varSelect
               );
            }

            // Perform the SQL query.
            ResultSet varResult = varDatabase.beginQuery(varSelect);

            // For each row in the account table, construct an
            // account object.
            while (varResult.next()) {
               LdbCardAccount varAccount = new LdbCardAccount(
                  varResult
               );
            }
         } catch (SQLException varException) {
            // If there was an error while reading the database,
            // log the error and abort.
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardAccount SQL Exception caught." +
                  "\n\tPRC = " + PRC_DATABASE_ERROR +
                  "\n\tSRC = " + SRC_ACCOUNT_SELECT_SQL_FAILURE
               );
            }
            ErrorLog.logError(
               RB_ID,
               MSG_ACCOUNT_SELECT_SQL_FAILURE,
               varException
            );
            throw new ETillAbortOperation(
               PRC_DATABASE_ERROR,
               SRC_ACCOUNT_SELECT_SQL_FAILURE
            );
         } finally {
            // Regardless of whether there was an exception or not,
            // clean up the ETillArchive object.
            try {
               varDatabase.endQuery();
            } catch (SQLException varException) {
               if (Trace.isAnyoneTracing()) {
                  Trace.traceErrorOccurred(
                     TRACE_ID,
                     "LdbCardAccount SQL Exception caught." +
                     "\n\tPRC = " + PRC_DATABASE_ERROR +
                     "\n\tSRC = " + SRC_ACCOUNT_SELECT_CLOSE_FAILURE
                  );
               }
               // If there was en error, it occured during cleanup so
               // there is no reason to abort.  Just log the error
               // because there is a defect in the ETillArchive object.
               ErrorLog.logError(
                  RB_ID,
                  MSG_ACCOUNT_SELECT_CLOSE_FAILURE,
                  varException
               );
            }
         }
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.resurrectAccounts()"
         );
      }
   }

   /********************************************************************
   * This method is used by the cassette when the framework asks the
   * cassette to verify the configuration.  In this implementation,
   * it is considered an error if the cassette was installed but
   * there are no accounts defined.
   *
   * @return - True if there is at least one account created during
   * initialization, False otherwise.
   ********************************************************************/
   public static boolean verifyAccounts() {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.verifyAccounts()"
         );
      }

      boolean varResult = true;

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.verifyAccounts()"
         );
      }
      return varResult;
   }

   /********************************************************************
   * This method is used by the cassette when the framework asks the
   * cassette to initialize.
   * <p>
   * For this cassette, there is nothing to do because account
   * initialization will continue when the AccountStart command is
   * processed.  Other cassettes may take this opportunity to
   * initialize objects associated with the cassette (e.g. brand or
   * currency objects)
   *
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public static void initializeAccounts()
      throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.initializeAccounts()"
         );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.initializeAccounts()"
         );
      }
   }

   /********************************************************************
   * This method is used by the cassette when the framework requests
   * an account object.  The requested account object is found in the
   * Hashtable and returned.
   * <p>
   * For this cassette, by definition, there can only be one account per
   * merchant, so the key required is the merchant number.  Other
   * cassettes may support multiple accounts per merchant and require
   * the account number as well as the merchant number to construct a
   * key.
   *
   * @param     argMerchant - The merchant number of the account object
   * to be returned.
   * @param     argAccountNumber - The account number of the account
   * object to be returned.
   * @return    - The account object for the specified merchant.
   ********************************************************************/
   public static LdbCardAccount retrieveAccount(
      String argMerchant
   ) {
      return (LdbCardAccount)clsAccountList.get(argMerchant);
   }

   /********************************************************************
   * This method is used by the cassette when validating an
   * adminstration request to create a new account object.
   * <p>
   * For this cassette, the protocol data values are validated.  Since
   * all the conversions have been performed successfully by the
   * framework (using the parameter validation table provided by the
   * cassette), this method merely validates that all required parameter
   * exist.
   * <p>
   * Note that this cassette assumes that there is only one account per
   * merchant.  If an administrator attempts to start another account
   * for the same merchant, even if the account number is different
   * (and therefore did not get automatically rejected by the framework)
   * it is considered a duplicate account and the request is rejected.
   *
   * @param     argRequest - The administration request that is
   * attempting to create a new account for this cassette.
   * @param     argResponse - The response object to be updated with
   * the result of the validation
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public static void validateCreateRequest(
      AdminRequest argRequest,
      APIResponse  argResponse
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.validateCreateRequest()"
         );
      }

      // Obtain parameter table that contains all the protocol data.
      ParameterTable varParameters = argRequest.getProtocolData();

      // Throw an exception if any required parameter is missing.
      // Note that if there is no protocol data parameter table at
      // all, it is treated the same as no currency specified.
      if (
         varParameters == null ||
         varParameters.get(PD_ACC_CURRENCY) == null
      ) {

         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardAccount Currency parameter is missing. " +
               "\n\tPRC = " + PRC_PARAMETER_NOT_FOUND +
               "\n\tSRC = " + SRC_CASSETTE_CURRENCY
             );
         }
         throw new ETillAbortOperation(
            PRC_PARAMETER_NOT_FOUND,
            SRC_CASSETTE_CURRENCY
         );

      } else if (varParameters.get(PD_ACC_USER) == null) {

         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardAccount User ID parameter is missing. " +
               "\n\tPRC = " + PRC_PARAMETER_NOT_FOUND +
               "\n\tSRC = " + SRC_CASSETTE_USER
             );
         }
         throw new ETillAbortOperation(
            PRC_PARAMETER_NOT_FOUND,
            SRC_CASSETTE_USER
         );

      } else if (varParameters.get(PD_ACC_PASSWORD) == null) {

         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardAccount Password parameter is missing. " +
               "\n\tPRC = " + PRC_PARAMETER_NOT_FOUND +
               "\n\tSRC = " + SRC_CASSETTE_PASSWORD
             );
         }
         throw new ETillAbortOperation(
            PRC_PARAMETER_NOT_FOUND,
            SRC_CASSETTE_PASSWORD
         );

      } else if (varParameters.get(PD_ACC_LDBCARDID) == null) {

         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardAccount Merchant's LdbCard ID is missing. " +
               "\n\tPRC = " + PRC_PARAMETER_NOT_FOUND +
               "\n\tSRC = " + SRC_CASSETTE_LDBCARDID
             );
         }
         throw new ETillAbortOperation(
            PRC_PARAMETER_NOT_FOUND,
            SRC_CASSETTE_LDBCARDID
         );

      } else if (varParameters.get(PD_ACC_URL) == null) {

         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardAccount RMI URL of payment engine is missing. " +
               "\n\tPRC = " + PRC_PARAMETER_NOT_FOUND +
               "\n\tSRC = " + SRC_CASSETTE_LDBCARDRMIURL
             );
         }
         throw new ETillAbortOperation(
            PRC_PARAMETER_NOT_FOUND,
            SRC_CASSETTE_LDBCARDRMIURL
         );

      }

      // Check to see if the specified currency is supported at
      // this installation.  If not, throw an exception.
      String varCurrency = (String)varParameters.get(PD_ACC_CURRENCY);
      String varExp10 = ErrorLog.lookupMessageText(
         RB_ID,
         varCurrency + ".Exp10"
      );
      String varLanguage = ErrorLog.lookupMessageText(
         RB_ID,
         varCurrency + ".Language"
      );
      if (Trace.isAnyoneTracing()) {
         Trace.traceDebug(
            TRACE_ID,
            "LdbCardAccount Currency values loaded. " +
            "\n\tCurrency    = " + varCurrency +
            "\n\tvarExp10    = " + varExp10 +
            "\n\tvarLanguage = " + varLanguage
          );
      }
      if (
         varExp10 == null || varLanguage == null ||
         varExp10.length() == 0 || varLanguage.length() == 0 ||
         varExp10.startsWith("Message text not found") ||
         varLanguage.startsWith("Message text not found")
      ) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardAccount Currency not supported. " +
               "\n\tCurrency    = " + varCurrency +
               "\n\tPRC         = " + PRC_PARAMETER_VALUE_ERROR +
               "\n\tSRC         = " + SRC_CASSETTE_CURRENCY
             );
         }

         throw new ETillAbortOperation(
            PRC_PARAMETER_VALUE_ERROR,
            SRC_CASSETTE_CURRENCY
         );

      }

      // If the hashtable already contains an account for this merchant,
      // this request must be aborted.  This cassette assumes that there
      // is only one account per merchant.
      ParameterTable varFrameworkKeywords =
         argRequest.getFrameworkKeywords();
      String varMerchant = (String)varFrameworkKeywords.get(MERCHANTNUMBER);
      String varAccount  = (String)varFrameworkKeywords.get(ACCOUNTNUMBER);
      if (clsAccountList.containsKey(varMerchant)) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardAccount Abort due to definition " +
               " of multiple accounts for the same merchant." +
               "\n\tPRC = " + PRC_DUPLICATE_OBJECT +
               "\n\tSRC = " + RC_ACCOUNT
            );
         }
         ErrorLog.logError(
            RB_ID,
            MSG_ACCOUNT_MULTIPLE_ACCOUNTS,
            null,
            varMerchant,
            ((LdbCardAccount)clsAccountList.get(varMerchant))
               .getAccountNumber(),
            varAccount
         );
         throw new ETillAbortOperation(
            PRC_DUPLICATE_OBJECT,
            RC_ACCOUNT
         );
      }

      // Indicate that the operation was a success.
      argResponse.setReturnCodes(
         PRC_OPERATION_SUCCESS,
         RC_NONE
      );

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.validateCreateRequest()"
         );
      }
   }

   /********************************************************************
   * This method is used by the cassette when validating an
   * adminstration request to modify an existing account object.
   * <p>
   * For this cassette, the protocol data values are validated.  Since
   * the conversions have been performed successfully by the framework
   * (using the parameter validation table provided by the cassette),
   * this method merely validates that the request does not attempt to
   * modify the currency value.  The existing orders that use this
   * account cannot change to a new currency simply because the
   * administrator has changed the account's currency.
   * <p>
   * All other account attributes are related to the payment engine
   * connection and can be changed without affecting the orders that
   * reference this account.
   *
   * @param     argRequest - The administration request that is
   * attempting to create a new account for this cassette.
   * @param     argResponse - The response object to be updated with
   * the result of the validation
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public static void validateModifyRequest(
      AdminRequest argRequest,
      APIResponse  argResponse
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.validateModifyRequest()"
         );
      }

      // Obtain parameter table that contains all the protocol data.
      ParameterTable varParameters = argRequest.getProtocolData();

      // Throw an exception if the administrator attempts to modify
      // the currency attribute.
      if (
         varParameters != null &&
         varParameters.get(PD_ACC_CURRENCY) != null
      ) {

         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardAccount Currency is not allowed. " +
               "\n\tPRC = " + PRC_PARAMETER_NOT_ALLOWED +
               "\n\tSRC = " + SRC_CASSETTE_CURRENCY
             );
         }
         throw new ETillAbortOperation(
            PRC_PARAMETER_NOT_ALLOWED,
            SRC_CASSETTE_CURRENCY
         );

      }

      // Indicate that the operation was a success.
      argResponse.setReturnCodes(
         PRC_OPERATION_SUCCESS,
         RC_NONE
      );

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.validateModifyRequest()"
         );
      }
   }

   //*******************************************************************
   //Instance data for the account object.
   //******************************************************************/
   /********************************************************************
   * This account attribute contains a reference to the frameworks
   * account object.
   ********************************************************************/
   private AccountAdmin objFrameworkAccount;
   /********************************************************************
   * This account attribute contains the merchant number.
   ********************************************************************/
   private String       objMerchant;
   /********************************************************************
   * This account attribute contains the account number.
   ********************************************************************/
   private String       objAccount;
   /********************************************************************
   * This account attribute contains the ISO currency code supported
   * by this account.
   ********************************************************************/
   private String       objCurrency;
   /********************************************************************
   * This account attribute contains the exponent on 10 used to
   * position the decimal point in integer currency values.
   ********************************************************************/
   private String       objExp10;
   /********************************************************************
   * This account attribute contains the ID of the user authorized
   * to connect to this account's RMI server.
   ********************************************************************/
   private String       objUser;
   /********************************************************************
   * This account attribute contains the password used to authenticate
   * the accounts user.
   ********************************************************************/
   private String       objPassword;
   /********************************************************************
   * This account attribute contains the merchant's identifier in the
   * payment engine.                               *
   ********************************************************************/
   private String       objLdbCardId;
   /********************************************************************
   * This account attribute contains the universal resource locator for
   * the RMI server used by this account.
   ********************************************************************/
   private String       objUrl;
   /********************************************************************
   * This account attribute contains the time of day when the cassette
   * should automatically close any open batch.  The value represents
   * the time of day in minutes after midnight.  The null value has
   * special meaning:  the cassette should never automatically close
   * an open batch at a particular time of day.
   ********************************************************************/
   private Integer      objBatchCloseTime;
   /********************************************************************
   * This account attribute contains the minimum batch total for batches
   * that are closed automatically at a particular time of day.
   * The null value has special meaning: any batch with any amount
   * larger than 0 should be closed when the configured time of day is
   * reached.
   ********************************************************************/
   private Integer      objMinBatchTotal;
   /********************************************************************
   * This account attribute contains the maximum batch size for batches
   * that use this account.  When the number of items in the batch
   * reach this value, the cassette will automatically close the batch.
   * The null value has special meaning: the cassette should never
   * use the batch size to determine if a batch should be closed
   * automatically.
   ********************************************************************/
   private Integer      objMaxBatchSize;
   /********************************************************************
   * This account attribute contains a number formatter configured to
   * convert currency values into strings.
   ********************************************************************/
   private NumberFormat objFormatter;
   /********************************************************************
   * This account attribute contains the exponent on 10 used to
   * position the decimal point in integer currency values.
   ********************************************************************/
   private double       objDoubleExp10;
   /********************************************************************
   * This account attribute contains the connection to the RMI server
   * used by this account.
   ********************************************************************/
   private Connection   objConnection;


   /********************************************************************
   * This method constructs a new account object and initializes the
   * object's data from the current row of the supplied ResultSet.
   *
   * @param     argCassette - The cassette object that is initializing.
   * @param     argResultSet - A ResultSet where the current row
   * contains the values to be used to initialize the new account
   * object.
   * @exception SQLException - any failure while reading the database
   * row that contains this object's definition results in this
   * exception.
   * @exception ETillAbortOperation - any other failure results in this
   * exception.
   ********************************************************************/
   protected LdbCardAccount(
      ResultSet        argResultSet
   ) throws SQLException, ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.LdbCardAccount()"
         );
      }
     this.objMerchant        =
        ETillArchive.readString(argResultSet, ACC_MERCHANT);
     this.objAccount         =
        ETillArchive.readString(argResultSet, ACC_ACCOUNT);
     this.objCurrency        =
        ETillArchive.readString(argResultSet, ACC_CURRENCY);
     this.objUser            =
        ETillArchive.readString(argResultSet, ACC_USER);
     this.objPassword        =
        ETillArchive.readString(argResultSet, ACC_PASSWORD);
     this.objLdbCardId       =
        ETillArchive.readString(argResultSet, ACC_LDBCARD_ID);
     this.objUrl             =
        ETillArchive.readString(argResultSet, ACC_URL);
     this.objBatchCloseTime  =
        ETillArchive.readInteger(argResultSet, ACC_CLOSE_TIME);
     this.objMinBatchTotal    =
        ETillArchive.readInteger(argResultSet, ACC_MIN_BATCH_TOTAL);
     this.objMaxBatchSize    =
        ETillArchive.readInteger(argResultSet, ACC_MAX_BATCH_SIZE);

     // Initialize the account's tools and add the account to the
     // list of cassette accounts.
     this.initializeAccount();

     if (Trace.isAnyoneTracing()) {
        Trace.traceFunctionExit(
           TRACE_ID,
           "LdbCardAccount.LdbCardAccount()"
        );
     }

   }

   /********************************************************************
   * This constructor is used by the cassette when the framework asks
   * it to service an adminitration request that creates a new account.
   * <p>
   * For this cassette, it initializes some attributes using values from
   * the protocol data table found in the request and then uses methods
   * common to all constructors to initialize the other account
   * attributes.
   *
   * @param     argRequest - the administration request that was sent
   * to the cassette to create a new account.
   * @param     argResponse - the response object to be updated with
   * the result of the operation.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public LdbCardAccount(
      AdminRequest  argRequest,
      APIResponse   argResponse
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.LdbCardAccount()"
         );
      }

      // Obtain parameter table that contains all the protocol data.
      ParameterTable varParameters = argRequest.getProtocolData();

      // Initialize this object from values in request and protocol data.
      this.objMerchant        =
         (String)argRequest.getValue(MERCHANTNUMBER);
      this.objAccount         =
         (String)argRequest.getValue(ACCOUNTNUMBER);
      this.objCurrency        =
         (String)varParameters.get(PD_ACC_CURRENCY);
      this.objUser            =
         (String)varParameters.get(PD_ACC_USER);
      this.objPassword        =
         (String)varParameters.get(PD_ACC_PASSWORD);
      this.objLdbCardId     =
         (String)varParameters.get(PD_ACC_LDBCARDID);
      this.objUrl             =
         (String)varParameters.get(PD_ACC_URL);
      this.objBatchCloseTime  =
         (Integer)varParameters.get(PD_ACC_BATCH_CLOSE_TIME, null);
      this.objMinBatchTotal    =
         (Integer)varParameters.get(PD_ACC_MIN_BATCH_TOTAL, null);
      this.objMaxBatchSize    =
         (Integer)varParameters.get(PD_ACC_MAX_BATCH_SIZE, null);

      // Initialize the account's tools and add the account to the
      // list of cassette accounts.
      this.initializeAccount();

      // Add this object to the create list.
      Supervisor.getThreadCommitPoint().addToCreateList(this);

      // Indicate that the operation was a success.
      argResponse.setReturnCodes(
         PRC_OPERATION_SUCCESS,
         RC_NONE
      );

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.LdbCardAccount()"
         );
      }
   }

   /********************************************************************
   * This method is used by all constructors to initialize the number
   * formatter used to convert integer ammounts into strings and to
   * add new accounts to the list of accounts kept by the account class.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   private void initializeAccount() throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.initializeAccount()"
         );
      }

      // Find the framework account object and save its reference
      // for convenience.
      this.objFrameworkAccount =
         Admin.getAccountAdminObject(
            CASSETTE_NAME,
            this.objMerchant,
            this.objAccount
         );

      // The payment engine requires amounts to be specified as
      // strings.  The Payment Manager amount is recorded as three
      // values: a binary integer, a binary ISO currency code, and
      // a binary exponent base 10.  To convert framework
      // amounts into strings, a NumberFormat object is needed.
      // To get the correct decimal character and currency character
      // in the result, the NumberFormat object is constructed using
      // a Locale object customized for the currency.  For more
      // information, see the convertAmount() method below.
      this.objExp10 = ErrorLog.lookupMessageText(
         RB_ID,
         this.objCurrency + ".Exp10"
      );
      String varLanguage = ErrorLog.lookupMessageText(
         RB_ID,
         this.objCurrency + ".Language"
      );
      String varCountry = ErrorLog.lookupMessageText(
         RB_ID,
         this.objCurrency + ".Country"
      );
      if (varCountry == null) {
         varCountry = "";
      }
      Locale varLocale = new Locale(varLanguage, varCountry);
      this.objFormatter = NumberFormat.getInstance(varLocale);
      int varExp10 = Integer.parseInt(this.objExp10);
      if (varExp10 < 0) {
         this.objFormatter.setMaximumFractionDigits(-varExp10);
         this.objFormatter.setMinimumFractionDigits(-varExp10);
      } else {
         this.objFormatter.setMaximumFractionDigits(0);
         this.objFormatter.setMinimumFractionDigits(0);
      }
      this.objDoubleExp10 = (double)varExp10;

      // Add this account to the list of accounts.
      this.addAccountToList();

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.initializeAccount()"
         );
      }
   }

   /********************************************************************
   * This method is used to add a new account to the list of accounts
   * kept by the account class.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   private void addAccountToList()
   throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.addAccountToList()"
         );
      }

      clsAccountList.put(
         this.getMerchant(),
         this
      );

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.addAccountToList()"
         );
      }
   }

   /********************************************************************
   * This method is used by the framework thread (a commitable object)
   * when its commit() method discovers the cassette account object on
   * the framework thread create list.
   *
   * @param     argConnection - The JDBC database connection to be used
   * for this operation.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void createRecord(
      java.sql.Connection    argConnection
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.createRecord()"
         );
      }

      try {
         // The SQL statement to insert a new cassette account.
         String varInsert =
            " INSERT INTO " +
            ETillArchive.geteTillDatabaseOwner() + "." + TBL_ACCOUNT + " (" +
            ACC_MERCHANT        + ", " +
            ACC_ACCOUNT         + ", " +
            ACC_CURRENCY        + ", " +
            ACC_USER            + ", " +
            ACC_PASSWORD        + ", " +
            ACC_LDBCARD_ID      + ", " +
            ACC_URL             + ", " +
            ACC_CLOSE_TIME      + ", " +
            ACC_MIN_BATCH_TOTAL + ", " +
            ACC_MAX_BATCH_SIZE  +
            ") VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)";
         if (Trace.isAnyoneTracing()) {
            Trace.traceDatabaseWrite(
               TRACE_ID,
               "\n" + varInsert + "\n" +
               ACC_MERCHANT   + " = " + this.getMerchant()      + "\n" +
               ACC_ACCOUNT    + " = " + this.getAccountNumber() + "\n" +
               ACC_CURRENCY   + " = " + this.getCurrency()      + "\n" +
               ACC_USER       + " = " + this.getUser()          + "\n" +
               ACC_PASSWORD   + " = " + this.getObscurePassword()
                                                                + "\n" +
               ACC_LDBCARD_ID + " = " + this.getLdbCardId()     + "\n" +
               ACC_URL        + " = " + this.getUrl()           + "\n" +
               ACC_CLOSE_TIME + " = " + this.getBatchCloseTimeObject()
                                                                + "\n" +
               ACC_MIN_BATCH_TOTAL + " = " + this.getMinBatchTotalObject()
                                                                + "\n" +
               ACC_MAX_BATCH_SIZE + " = " + this.getMaxBatchSizeObject()
            );
         }
         // Ask the connection to build a prepared statement using
         // the SQL statement defined above.
         PreparedStatement varStatement =
            argConnection.prepareStatement(varInsert);
         // Clear all the variable parameters from the statement.
         varStatement.clearParameters();
         // Tell the statement to set the variable values.
         ETillArchive.addString(this.getMerchant(),      varStatement, 1);
         ETillArchive.addString(this.getAccountNumber(), varStatement, 2);
         ETillArchive.addString(this.getCurrency(),      varStatement, 3);
         ETillArchive.addString(this.getUser(),          varStatement, 4);
         ETillArchive.addString(this.getPassword(),      varStatement, 5);
         ETillArchive.addString(this.getLdbCardId(),     varStatement, 6);
         ETillArchive.addString(this.getUrl(),           varStatement, 7);
         ETillArchive.addInteger(this.getBatchCloseTimeObject(),
                                                         varStatement, 8);
         ETillArchive.addInteger(this.getMinBatchTotalObject(),
                                                         varStatement, 9);
         ETillArchive.addInteger(this.getMaxBatchSizeObject(),
                                                         varStatement, 10);
         // Execute the insert.
         int varRowsAffected = varStatement.executeUpdate();
         // Close the statement.
         varStatement.close();
         // Assertion test:  If the number of affected rows is not 1
         // there is a serious defect and the cassette cannot continue.
         // The error is logged and a cassette exception is thrown.
         if (varRowsAffected != 1) {
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardAccount Created too many rows." +
                  "\n\tPRC = " + PRC_DATABASE_ERROR +
                  "\n\tSRC = " + SRC_ACCOUNT_CREATE_ROW_FAILURE
               );
            }
            Vector varVector = new Vector();
            varVector.addElement(this.getMerchant());
            varVector.addElement(this.getAccountNumber());
            varVector.addElement(Integer.toString(varRowsAffected));
            ErrorLog.log(
               RB_ID,
               MSG_ACCOUNT_CREATE_ROW_FAILURE,
               varVector
            );
            throw new ETillCassetteException(
               ErrorLog.lookupMessage(
                  RB_ID,
                  MSG_ACCOUNT_CREATE_ROW_FAILURE,
                  varVector
               )
            );
         }
      }  catch (SQLException varException) {
         // The update failed.  Log the error and abort.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardAccount SQLException caught." +
               "\n\tPRC = " + PRC_DATABASE_ERROR +
               "\n\tSRC = " + SRC_ACCOUNT_CREATE_SQL_FAILURE
            );
         }
         ErrorLog.logError(
            RB_ID,
            MSG_ACCOUNT_CREATE_SQL_FAILURE,
            varException,
            this.getMerchant(),
            this.getAccountNumber()
         );
         throw new ETillAbortOperation(
            PRC_DATABASE_ERROR,
            SRC_ACCOUNT_CREATE_SQL_FAILURE
         );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.createRecord()"
         );
      }
   }

   /********************************************************************
   * This method is used by the framework thread (a commitable object)
   * when its commit() method discovers the cassette account object on
   * the framework thread update list.
   *
   * @param     argConnection - The JDBC database connection to be used
   * for this operation.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void updateRecord(
      java.sql.Connection    argConnection
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.updateRecord()"
         );
      }

      try {
         // The SQL statement to update an existing cassette account.
         String varUpdate =
            " UPDATE " +
            ETillArchive.geteTillDatabaseOwner() + "." + TBL_ACCOUNT +
            " SET " +
            ACC_CURRENCY        + "=?, " +
            ACC_USER            + "=?, " +
            ACC_PASSWORD        + "=?, " +
            ACC_LDBCARD_ID      + "=?, " +
            ACC_URL             + "=?, " +
            ACC_CLOSE_TIME      + "=?, " +
            ACC_MIN_BATCH_TOTAL + "=?, " +
            ACC_MAX_BATCH_SIZE  + "=?  " +
            " WHERE " +
            ACC_MERCHANT        + "= ? AND " +
            ACC_ACCOUNT         + "= ?";
         if (Trace.isAnyoneTracing()) {
            Trace.traceDatabaseWrite(
               TRACE_ID,
               "\n" + varUpdate + "\n" +
               ACC_CURRENCY   + " = " + this.getCurrency()      + "\n" +
               ACC_USER       + " = " + this.getUser()          + "\n" +
               ACC_PASSWORD   + " = " + this.getObscurePassword()
                                                                + "\n" +
               ACC_LDBCARD_ID + " = " + this.getLdbCardId()     + "\n" +
               ACC_URL        + " = " + this.getUrl()           + "\n" +
               ACC_CLOSE_TIME + " = " + this.getBatchCloseTimeObject()
                                                                + "\n" +
               ACC_MIN_BATCH_TOTAL + " = " + this.getMinBatchTotalObject()
                                                                + "\n" +
               ACC_MAX_BATCH_SIZE + " = " + this.getMaxBatchSizeObject()
                                                                + "\n" +
               ACC_MERCHANT   + " = " + this.getMerchant()      + "\n" +
               ACC_ACCOUNT    + " = " + this.getAccountNumber()
            );
         }
         // Ask the connection to build a prepared statement using
         // the SQL statement defined above.
         PreparedStatement varStatement =
            argConnection.prepareStatement(varUpdate);
         // Clear all the variable parameters from the statement.
         varStatement.clearParameters();
         // Tell the statement to set the variable values.
         ETillArchive.addString(this.getCurrency(),      varStatement, 1);
         ETillArchive.addString(this.getUser(),          varStatement, 2);
         ETillArchive.addString(this.getPassword(),      varStatement, 3);
         ETillArchive.addString(this.getLdbCardId(),     varStatement, 4);
         ETillArchive.addString(this.getUrl(),           varStatement, 5);
         ETillArchive.addInteger(this.getBatchCloseTimeObject(),
                                                         varStatement, 6);
         ETillArchive.addInteger(this.getMinBatchTotalObject(),
                                                         varStatement, 7);
         ETillArchive.addInteger(this.getMaxBatchSizeObject(),
                                                         varStatement, 8);
         ETillArchive.addString(this.getMerchant(),      varStatement, 9);
         ETillArchive.addString(this.getAccountNumber(), varStatement, 10);
         // Execute the update.
         int varRowsAffected = varStatement.executeUpdate();
         // Close the statement.
         varStatement.close();
         // Assertion test:  If the number of affected rows is not 1
         // there is a serious defect and the cassette cannot continue.
         // The error is logged and a cassette exception is thrown.
         if (varRowsAffected != 1) {
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardAccount Modified too many rows." +
                  "\n\tPRC = " + PRC_DATABASE_ERROR +
                  "\n\tSRC = " + SRC_ACCOUNT_UPDATE_ROW_FAILURE
               );
            }
            Vector varVector = new Vector();
            varVector.addElement(this.getMerchant());
            varVector.addElement(this.getAccountNumber());
            varVector.addElement(Integer.toString(varRowsAffected));
            ErrorLog.log(
               RB_ID,
               MSG_ACCOUNT_UPDATE_ROW_FAILURE,
               varVector
            );
            throw new ETillCassetteException(
               ErrorLog.lookupMessage(
                  RB_ID,
                  MSG_ACCOUNT_UPDATE_ROW_FAILURE,
                  varVector
               )
            );
         }
      }  catch (SQLException varException) {
         // The update failed.  Log the error and abort.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardAccount SQLException caught." +
               "\n\tPRC = " + PRC_DATABASE_ERROR +
               "\n\tSRC = " + SRC_ACCOUNT_UPDATE_SQL_FAILURE
            );
         }
         ErrorLog.logError(
            RB_ID,
            MSG_ACCOUNT_UPDATE_SQL_FAILURE,
            varException,
            this.getMerchant(),
            this.getAccountNumber()
         );
         throw new ETillAbortOperation(
            PRC_DATABASE_ERROR,
            SRC_ACCOUNT_UPDATE_SQL_FAILURE
         );
      }


      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.updateRecord()"
         );
      }
   }

   /********************************************************************
   * This method is used by the framework thread (a commitable object)
   * when its commit() method discovers the cassette account object on
   * the framework thread delete list.
   *
   * @param     argConnection - The JDBC database connection to be used
   * for this operation.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void deleteRecord(
      java.sql.Connection    argConnection
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.deleteRecord()"
         );
      }

      try {
         // The SQL statement to delete an existing cassette account.
         String varDelete =
            " DELETE FROM " +
            ETillArchive.geteTillDatabaseOwner() + "." + TBL_ACCOUNT +
            " WHERE " +
            ACC_MERCHANT     + "= ? AND " +
            ACC_ACCOUNT      + "= ?";
         if (Trace.isAnyoneTracing()) {
            Trace.traceDatabaseWrite(
               TRACE_ID,
               "\n" + varDelete + "\n" +
               ACC_MERCHANT   + " = " + this.getMerchant()      + "\n" +
               ACC_ACCOUNT    + " = " + this.getAccountNumber()
            );
         }
         // Ask the connection to build a prepared statement using
         // the SQL statement defined above.
         PreparedStatement varStatement =
            argConnection.prepareStatement(varDelete);
         // Clear all the variable parameters from the statement.
         varStatement.clearParameters();
         // Tell the statement to set the variable values.
         ETillArchive.addString(this.getMerchant(),      varStatement, 1);
         ETillArchive.addString(this.getAccountNumber(), varStatement, 2);
         // Execute the delete.
         int varRowsAffected = varStatement.executeUpdate();
         // Close the statement.
         varStatement.close();
         // Assertion test:  If the number of affected rows is not 1
         // there is a serious defect and the cassette cannot continue.
         // The error is logged and a cassette exception is thrown.
         if (varRowsAffected != 1) {
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardAccount Deleted too many rows." +
                  "\n\tPRC = " + PRC_DATABASE_ERROR +
                  "\n\tSRC = " + SRC_ACCOUNT_DELETE_ROW_FAILURE
               );
            }
            Vector varVector = new Vector();
            varVector.addElement(this.getMerchant());
            varVector.addElement(this.getAccountNumber());
            varVector.addElement(Integer.toString(varRowsAffected));
            ErrorLog.log(
               RB_ID,
               MSG_ACCOUNT_DELETE_ROW_FAILURE,
               varVector
            );
            throw new ETillCassetteException(
               ErrorLog.lookupMessage(
                  RB_ID,
                  MSG_ACCOUNT_DELETE_ROW_FAILURE,
                  varVector
               )
            );
         }
      }  catch (SQLException varException) {
         // The delete failed.  Log the error and abort.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardAccount SQLException caught." +
               "\n\tPRC = " + PRC_DATABASE_ERROR +
               "\n\tSRC = " + SRC_ACCOUNT_DELETE_SQL_FAILURE
            );
         }
         ErrorLog.logError(
            RB_ID,
            MSG_ACCOUNT_DELETE_SQL_FAILURE,
            varException,
            this.getMerchant(),
            this.getAccountNumber()
         );
         throw new ETillAbortOperation(
            PRC_DATABASE_ERROR,
            SRC_ACCOUNT_DELETE_SQL_FAILURE
         );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.deleteRecord()"
         );
      }
   }

   /********************************************************************
   * This method is used by the cassette when the framework asks
   * it to service an adminitration request that modifies an
   * existing account.
   * <p>
   * For this cassette, it re-initializes its attributes using values
   * from the protocol data table found in the request and adds the
   * object to the update list of the current commit point.
   *
   * @param     argRequest - the administration request that was sent
   * to the cassette to modify the account.
   * @param     argResponse - the response object to be updated with
   * the result of the operation.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void modifyAccount(
      AdminRequest  argRequest,
      APIResponse   argResponse
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.modifyAccount()"
         );
      }

      // Obtain parameter table that contains all the protocol data.
      ParameterTable varParameters = argRequest.getProtocolData();

      // Make sure cassette dependent data is being modified.
      if (varParameters != null) {

         // There are valid ModifyAccount commands that do not modify
         // any cassette account attributes.  Most notably, an
         // administrator can disable (stop) an account with changing
         // any attributes.  The following test assures that some
         // attribute is to be changed before it attempts to update the
         // account, add the account to the commit point, and set the
         // framework pending flag.
         if (
            varParameters.containsKey(PD_ACC_USER) ||
            varParameters.containsKey(PD_ACC_PASSWORD) ||
            varParameters.containsKey(PD_ACC_LDBCARDID) ||
            varParameters.containsKey(PD_ACC_URL) ||
            varParameters.containsKey(PD_ACC_BATCH_CLOSE_TIME) ||
            varParameters.containsKey(PD_ACC_MIN_BATCH_TOTAL) ||
            varParameters.containsKey(PD_ACC_MAX_BATCH_SIZE)
         ) {
            // Initialize this object from values in protocol data.  Note
            // that for any value that is not specified, the value will
            // not be changed because the default value is set to the
            // current value.
            this.objUser            =
               (String)varParameters.get(PD_ACC_USER, this.objUser);
            this.objPassword        =
               (String)varParameters.get(PD_ACC_PASSWORD, this.objPassword);
            this.objLdbCardId     =
               (String)varParameters.get(PD_ACC_LDBCARDID, this.objLdbCardId);
            this.objUrl             =
               (String)varParameters.get(PD_ACC_URL, this.objUrl);
            this.objBatchCloseTime  =
               (Integer)varParameters.get(PD_ACC_BATCH_CLOSE_TIME, this.objBatchCloseTime);
            this.objMinBatchTotal    =
               (Integer)varParameters.get(PD_ACC_MIN_BATCH_TOTAL, this.objMinBatchTotal);
            this.objMaxBatchSize    =
               (Integer)varParameters.get(PD_ACC_MAX_BATCH_SIZE, this.objMaxBatchSize);

            Supervisor.getThreadCommitPoint().addToUpdateList(this);

            // Indicate that this object has been modified and the
            // changes will not take effect until the account is
            // started.  Although the instance data has been modified
            // and the database will be modified when the database update
            // is committed, the RMI connection that uses these values
            // will not be stopped and recreated until the account is
            // stopped and re-started.  The pending flag is used to
            // to record this fact.
            this.objFrameworkAccount.setPending(true);
         }
      }

      // Indicate that the operation was a success.
      argResponse.setReturnCodes(
         PRC_OPERATION_SUCCESS,
         RC_NONE
      );

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.modifyAccount()"
         );
      }
   }

   /********************************************************************
   * This method is used by the cassette when the framework asks
   * it to service an adminitration request that deletes an
   * existing account.
   * <p>
   * For this cassette, it adds this object to the delete list of the
   * current commit point.
   *
   * @param     argRequest - the administration request that was sent
   * to the cassette to delete the account.
   * @param     argResponse - the response object to be updated with
   * the result of the operation.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void deleteAccount(
      AdminRequest  argRequest,
      APIResponse   argResponse
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.deleteAccount()"
         );
      }

      // Remove the object from the list of available accounts.
      clsAccountList.remove(
         this.objMerchant
      );

      // Delete the account from persistent storage.
      Supervisor.getThreadCommitPoint().addToDeleteList(this);

      // Indicate that the operation was a success.
      argResponse.setReturnCodes(
         PRC_OPERATION_SUCCESS,
         RC_NONE
      );

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.deleteAccount()"
         );
      }
   }

   /********************************************************************
   * This method is used by the cassette when the framework asks
   * it to service an adminitration request that starts an
   * existing account.
   * <p>
   * For this cassette, it creates a connection to the payment engine
   * and then starts the service threads for any batch associated
   * with this account.
   *
   * @param     argRequest - the administration request that was sent
   * to the cassette to start the account.
   * @param     argResponse - the response object to be updated with
   * the result of the operation.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void startAccount(
      AdminRequest  argRequest,
      APIResponse   argResponse
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.startAccount()"
         );
      }

      // Connect the account to the payment engine.
      this.connect();

      // Since the cassette batch object cannot initialize its
      // own service threads when it is constructed, it is
      // necessary complete the initialization here for each cassette
      // batch object that was resurrected.  The service threads
      // require the framework batch to be fully initialized so that
      // the BatchCloseRequest can be constructed and the framework
      // batch object is not initialized when the cassette batch is
      // constructed.
      Hashtable varBatchCollection =
         this.objFrameworkAccount.batchesForAccount();
      Enumeration varBatchList = varBatchCollection.elements();
      while (varBatchList.hasMoreElements()) {
         Batch varFrameworkBatch = (Batch)varBatchList.nextElement();
         LdbCardBatch varCassetteBatch =
            (LdbCardBatch)varFrameworkBatch.getCassetteBatch();
         varCassetteBatch.initializeServiceThreadsIfNecessary();
      }

      // Indicate that the operation was a success.
      argResponse.setReturnCodes(
         PRC_OPERATION_SUCCESS,
         RC_NONE
      );

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.startAccount()"
         );
      }
   }

   /********************************************************************
   * This method is used by the cassette when the framework asks
   * it to service an adminitration request that stops an
   * existing account.
   * <p>
   * For this cassette, it breaks a connection to the payment engine.
   *
   * @param     argRequest - the administration request that was sent
   * to the cassette to start the account.
   * @param     argResponse - the response object to be updated with
   * the result of the operation.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void stopAccount(
      AdminRequest  argRequest,
      APIResponse   argResponse
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.stopAccount()"
         );
      }

      // Disconnect the account from payment engine.
      this.disconnect();

      // Indicate that the operation was a success.
      argResponse.setReturnCodes(
         PRC_OPERATION_SUCCESS,
         RC_NONE
      );

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.stopAccount()"
         );
      }
   }

   /********************************************************************
   * This method is used by payment objects while processing a Deposit
   * command.  It finds the appropriate batch and adds the payment
   * to the batch.
   * <p>
   * The method must be synchronized because framework API locks do not
   * prevent race conditions between automatic batch close requests
   * and API requests that affect the contents of the batch (Deposit,
   * DepositReversal, Refund and RefundReversal).  Synchronizing on the
   * cassette account object prevents one thread from changing the
   * contents of the batch a micro-moment after some other thread has
   * triggered an automatic batch close.
   *
   * @param     argFrameworkPayment - a framework payment that is to
   * be added to a batch.
   * @exception ETillAbortOperation - any error results in this
   * exception
   ********************************************************************/
   public synchronized void selectBatchAndAddPayment(
      Payment argFrameworkPayment
   ) throws ETillAbortOperation {
      Batch varFrameworkBatch = this.selectBatch();
      LdbCardBatch varCassetteBatch =
         (LdbCardBatch)varFrameworkBatch.getCassetteBatch();
      varCassetteBatch.addPayment(argFrameworkPayment);
   }

   /********************************************************************
   * This method is used by credit objects while processing a Refund
   * command.  It finds the appropriate batch and adds the credit
   * to the batch.
   * <p>
   * The method must be synchronized because framework API locks do not
   * prevent race conditions between automatic batch close requests
   * and API requests that affect the contents of the batch (Deposit,
   * DepositReversal, Refund and RefundReversal).  Synchronizing on the
   * cassette account object prevents one thread from changing the
   * contents of the batch a micro-moment after some other thread has
   * triggered an automatic batch close.
   *
   * @param     argFrameworkCredit - a framework credit that is to
   * be added to a batch.
   * @exception ETillAbortOperation - any error results in this
   * exception
   ********************************************************************/
   public synchronized void selectBatchAndAddCredit(
      Credit argFrameworkCredit
   ) throws ETillAbortOperation {
      Batch varFrameworkBatch = this.selectBatch();
      LdbCardBatch varCassetteBatch =
         (LdbCardBatch)varFrameworkBatch.getCassetteBatch();
      varCassetteBatch.addCredit(argFrameworkCredit);
   }

   /********************************************************************
   * This method is used by payment objects while processing a
   * DepositReversal command.  It retrieves the batch that contains the
   * the payment and removes the payment from the batch.
   * <p>
   * The method must be synchronized because framework API locks do not
   * prevent race conditions between automatic batch close requests
   * and API requests that affect the contents of the batch (Deposit,
   * DepositReversal, Refund and RefundReversal).  Synchronizing on the
   * cassette account object prevents one thread from changing the
   * contents of the batch a micro-moment after some other thread has
   * triggered an automatic batch close.
   *
   * @param     argFrameworkPayment - a framework payment that is to
   * be removed from its batch.
   * @exception ETillAbortOperation - any error results in this
   * exception
   ********************************************************************/
   public synchronized void retrieveBatchAndRemovePayment(
      Payment argFrameworkPayment
   ) throws ETillAbortOperation {
      Batch varFrameworkBatch = this.retrieveBatch(argFrameworkPayment);
      LdbCardBatch varCassetteBatch =
         (LdbCardBatch)varFrameworkBatch.getCassetteBatch();
      varCassetteBatch.removePayment(argFrameworkPayment);
   }

   /********************************************************************
   * This method is used by credit objects while processing a
   * RefundReversal command.  It retrieves the batch that contains the
   * the credit and removes the credit from the batch.
   * <p>
   * The method must be synchronized because framework API locks do not
   * prevent race conditions between automatic batch close requests
   * and API requests that affect the contents of the batch (Deposit,
   * DepositReversal, Refund and RefundReversal).  Synchronizing on the
   * cassette account object prevents one thread from changing the
   * contents of the batch a micro-moment after some other thread has
   * triggered an automatic batch close.
   *
   * @param     argFrameworkCredit - a framework credit that is to
   * be removed from its batch.
   * @exception ETillAbortOperation - any error results in this
   * exception
   ********************************************************************/
   public synchronized void retrieveBatchAndRemoveCredit(
      Credit argFrameworkCredit
   ) throws ETillAbortOperation {
      Batch varFrameworkBatch = this.retrieveBatch(argFrameworkCredit);
      LdbCardBatch varCassetteBatch =
         (LdbCardBatch)varFrameworkBatch.getCassetteBatch();
      varCassetteBatch.removeCredit(argFrameworkCredit);
   }

   /********************************************************************
   * This method is used by payment objects to obtain the batch object
   * that will contain the payment when the payment is deposited.  It
   * is also used by credit objects to obtain the batch object that
   * will contain the credit when the credit is refunded.
   * <p>
   * The method must be synchronized because framework API locks do not
   * prevent race conditions between automatic batch open requests
   * and API requests that add a transaction to a batch (Deposit and
   * Refund).  Synchronizing on the cassette account object prevents
   * one thread from triggering an automatic batch open when no suitable
   * batch was found a micro-moment after some other thread has
   * triggered a batch open request for the same reason.
   *
   * @return    - the framework batch object of the current batch.
   * @exception ETillAbortOperation - any error results in this
   * exception
   ********************************************************************/
   private synchronized Batch selectBatch(
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.selectBatch()"
         );
      }
      Batch varResult = null;

      // Search for a suitable batch.  A collection of open batches
      // is obtained from the framework account object.  Some of those
      // open batches may be closed to new transactions because a
      // BatchClose request failed (typically due to a balance failure).
      // This logic will search for a batch that is open and NOT
      // closed to new transactions.
      Hashtable varBatchCollection =
         this.objFrameworkAccount.batchesForAccount();
      int varNumberOfBatchesFound = 0;
      Enumeration varBatchList = varBatchCollection.elements();
      while (varBatchList.hasMoreElements()) {
         Batch varFrameworkBatch = (Batch)varBatchList.nextElement();
         LdbCardBatch varCassetteBatch =
            (LdbCardBatch)varFrameworkBatch.getCassetteBatch();
         if (!varCassetteBatch.isBatchClosedToNewTransactions()) {
            varResult = varFrameworkBatch;
            varNumberOfBatchesFound++;
         }
      }

      // Assertion Test:  If there is more than one batch that is
      // open for new transactions, there is a serious defect in
      // the cassette.
      if (varNumberOfBatchesFound > 1) {

         // Assertion failure.  There should never be more than one
         // open batch that is also open to new transactions.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardAccount More than one batch open" +
               " for new transactions at the same time." +
               "\n\tPRC = " + PRC_CASSETTE_ERROR +
               "\n\tSRC = " + SRC_ACCOUNT_MORE_THAN_ONE_BATCH
            );
         }
         Vector varVector = new Vector();
         varVector.addElement(this.getMerchant());
         varVector.addElement(this.getAccountNumber());
         ErrorLog.log(
            RB_ID,
            MSG_ACCOUNT_MORE_THAN_ONE_BATCH,
            varVector
         );
         throw new ETillCassetteException(
            ErrorLog.lookupMessage(
               RB_ID,
               MSG_ACCOUNT_MORE_THAN_ONE_BATCH,
               varVector
            )
         );

      }

      // Check to see if we failed to find a suitable batch.
      if (varResult == null) {

         // No suitable batch was found.  Ask the Supervisor to
         // create a new batch, initialize it, and return it to the
         // caller.
         varResult = Supervisor.createBatch(
            this.getMerchant(),
            this.getAccountNumber(),
            PAYMENT_TYPE
         );
         varResult.setCurrentState(BATCH_OPEN);
         // Since the cassette batch object cannot initialize its
         // own service threads when it is constructed, it is
         // necessary complete the initialization here.  Service
         // threads require the framework batch to be fully initialized
         // so that the BatchCloseRequest can be constructed and
         // the framework batch object is not initialized when the
         // cassette batch is constructed.
         LdbCardBatch varCassetteBatch =
            (LdbCardBatch)varResult.getCassetteBatch();
         varCassetteBatch.initializeServiceThreadsIfNecessary();
         Supervisor.getThreadCommitPoint().addToCreateList(varResult);

      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.selectBatch()"
         );
      }
      return varResult;
   }

   /********************************************************************
   * This method is used by payment objects to obtain the batch object
   * that currently contains the payment.  It is also used by credit
   * objects to obtain the batch object that currently contains the
   * credit.
   *
   * @param     argFrameworkTransaction - a framework transaction (it
   * dosen't matter if it is a framework payment or a framework
   * credit) that is contained in the batch to be retrieved
   * @return    - the framework batch object of the current batch.
   * @exception ETillAbortOperation - any error results in this
   * exception
   ********************************************************************/
   private Batch retrieveBatch(
      Transaction argFrameworkTransaction
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.retrieveBatch()"
         );
      }
      String varBatchNumber = argFrameworkTransaction.getBatchNumber();
      // Assertion test:  This method should never be called before
      // the transaction has been placed in a batch.  If the batch
      // number is null, there is a serious defect and the cassette
      // cannot continue.
      if (varBatchNumber == null) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardAccount Attempting to retrieve a" +
               " batch for a transaction with a null batch number." +
               "\n\tPRC = " + PRC_CASSETTE_ERROR +
               "\n\tSRC = " + SRC_ACCOUNT_NULL_BATCH_NUMBER
            );
         }
         Vector varVector = new Vector();
         varVector.addElement(this.getMerchant());
         varVector.addElement(this.getAccountNumber());
         varVector.addElement(
            argFrameworkTransaction.getOrder().getOrderNumber()
         );
         varVector.addElement(
            argFrameworkTransaction.getTransactionNumber()
         );
         ErrorLog.log(
            RB_ID,
            MSG_ACCOUNT_NULL_BATCH_NUMBER,
            varVector
         );
         throw new ETillCassetteException(
            ErrorLog.lookupMessage(
               RB_ID,
               MSG_ACCOUNT_NULL_BATCH_NUMBER,
               varVector
            )
         );
      }
      Batch varResult = Supervisor.retrieveBatch(
         this.getMerchant(),
         varBatchNumber
      );

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.retrieveBatch()"
         );
      }
      return varResult;
   }

   /********************************************************************
   * This method is used to get the merchant number of this account.
   *
   * @return   - The merchant number.
   ********************************************************************/
   public String getMerchant() {
      return this.objMerchant;
   }

   /********************************************************************
   * This method is used to get the account number of this account.
   *
   * @return   - The account number.
   ********************************************************************/
   public String getAccountNumber() {
      return this.objAccount;
   }

   /********************************************************************
   * This method is used to get the ISO currency code used by this
   * account.
   *
   * @return   - The ISO currency code.
   ********************************************************************/
   public String getCurrency() {
      return this.objCurrency;
   }

   /********************************************************************
   * This method is used to get the user ID used for the Java RMI
   * connection.
   *
   * @return   - The user ID.
   ********************************************************************/
   public String getUser() {
      return this.objUser;
   }

   /********************************************************************
   * This method is used to get the password necessary to authenticate
   * the user ID.
   *
   * @return   - The user's password.
   ********************************************************************/
   public String getPassword() {
      return this.objPassword;
   }

   /********************************************************************
   * This method is used to get the password necessary to authenticate
   * the user ID when the password will be displayed in a trace.
   *
   * @return   - A string of asterisks the same length as the password.
   ********************************************************************/
   public String getObscurePassword() {
      StringBuffer varBuffer = new StringBuffer();
      for (
         int varIndex = 0;
         varIndex < this.objPassword.length();
         varIndex++
      ) {
         varBuffer.append("*");
      }
      return varBuffer.toString();
   }

   /********************************************************************
   * This method is used to get the identifier configured in the
   * the payment engine.
   *
   * @return   - The ID used by the payment engine to identify the
   * merchant using the account.
   ********************************************************************/
   public String getLdbCardId() {
      return this.objLdbCardId;
   }

   /********************************************************************
   * This method is used to get the URL used to locate the Java RMI
   * server for the payment engine.
   *
   * @return    - The URL that locates the RMI server.
   ********************************************************************/
   public String getUrl() {
      return this.objUrl;
   }

   /********************************************************************
   * This method is used to determine if the administrator has
   * configured the account to automatically close an open batch at
   * a specified time of day.  If it returns false, the automatic
   * time of day close should not be performed.  If it returns true,
   * the getBatchCloseTime() method should be used to obtain the
   * time of day specified in minutes after midnight.
   *
   * @return    - True if this account has been configured to
   * automatically close an open batch at a specified time of day,
   * false otherwise.
   ********************************************************************/
   public boolean isBatchCloseTimeConfigured() {
      boolean varResult = true;
      if (this.objBatchCloseTime == null) {
         varResult = false;
      }
      return varResult;
   }

   /********************************************************************
   * This method is used to get the time of day, specified in minutes
   * after midnight, when this account's open batches should be
   * automatically closed by the cassette.
   *
   * @return    - The number of minutes after midnight when an open
   * batch should be automatically closed.
   ********************************************************************/
   public int getBatchCloseTime() {

      // Assertion test:  If the batch close time has not been
      // configured, the value returned by this method is undefined.
      // It is considered a serious defect if this method is called
      // when the batch close time is not configured.  The error
      // is logged and a cassette exception is thrown.
      if (!this.isBatchCloseTimeConfigured()) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardAccount Attempting to get BatchCloseTime" +
               " but BatchCloseTime is not Configured." +
               "\n\tPRC = " + PRC_CASSETTE_ERROR +
               "\n\tSRC = " + SRC_ACCOUNT_NULL_BATCH_CLOSE_TIME
            );
         }
         Vector varVector = new Vector();
         varVector.addElement(this.getMerchant());
         varVector.addElement(this.getAccountNumber());
         ErrorLog.log(
            RB_ID,
            MSG_ACCOUNT_NULL_BATCH_CLOSE_TIME,
            varVector
         );
         throw new ETillCassetteException(
            ErrorLog.lookupMessage(
               RB_ID,
               MSG_ACCOUNT_NULL_BATCH_CLOSE_TIME,
               varVector
            )
         );
      }
      return this.objBatchCloseTime.intValue();
   }

   /********************************************************************
   * This method is used to get an object that represents the time of
   * day (specified in minutes after midnight) when this account's
   * open batches will be automatically closed.
   *
   * @return    - Either null or a reference to an Integer object with
   * a value equal to the configured batch close time.
   ********************************************************************/
   public Integer getBatchCloseTimeObject () {
      return this.objBatchCloseTime;
   }

   /********************************************************************
   * This method is used to determine if the administrator has
   * configured a minimum batch size for a batch that is closed at
   * a specified time of day.  If it returns false, no minimum batch
   * size was specified, so an automatic close should be performed if
   * there are one or more items in the batch.  If it returns true,
   * the getMinBatchTotal() method should be used to obtain the
   * minimum batch size and, if the batch size is smaller, the
   * automatic time of day close should be aborted.
   *
   * @return    - True if a minimum batch size has been configured,
   * false otherwise.
   ********************************************************************/
   public boolean isMinBatchTotalConfigured() {
      boolean varResult = true;
      if (this.objMinBatchTotal == null) {
         varResult = false;
      }
      return varResult;
   }

   /********************************************************************
   * This method is used to get minimum batch size allowed when the
   * batch is closed automatically.  If the number of items in the
   * batch is less than the minimum at the given time of day, the
   * automatic close will be abandoned.
   *
   * @return    - The minimum number of batch items that must exist
   * in an open batch before it is automatically closed at the given
   * time of day.
   ********************************************************************/
   public int getMinBatchTotal() {

      // Assertion test:  If the minimum batch size has not been
      // configured, the value returned by this method is undefined.
      // It is considered a serious defect if this method is called
      // when the minimum batch size is not configured.  The error
      // is logged and a cassette exception is thrown.
      if (!this.isMinBatchTotalConfigured()) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardAccount Attempting to get MinBatchTotal" +
               " but MinBatchTotal is not Configured." +
               "\n\tPRC = " + PRC_CASSETTE_ERROR +
               "\n\tSRC = " + SRC_ACCOUNT_NULL_MIN_BATCH_TOTAL
            );
         }
         Vector varVector = new Vector();
         varVector.addElement(this.getMerchant());
         varVector.addElement(this.getAccountNumber());
         ErrorLog.log(
            RB_ID,
            MSG_ACCOUNT_NULL_MIN_BATCH_TOTAL,
            varVector
         );
         throw new ETillCassetteException(
            ErrorLog.lookupMessage(
               RB_ID,
               MSG_ACCOUNT_NULL_MIN_BATCH_TOTAL,
               varVector
            )
         );
      }
      return this.objMinBatchTotal.intValue();
   }

   /********************************************************************
   * This method is used to get an object that represents the minimum
   * batch size for batches are that automatically closed at a given
   * time of day.
   *
   * @return    - Either null or a reference to an Integer object with
   * a value equal to the configured minimum batch size.
   ********************************************************************/
   public Integer getMinBatchTotalObject () {
      return this.objMinBatchTotal;
   }


   /********************************************************************
   * This method is used to determine if the administrator has
   * configured the account to automatically close an open batch when
   * that batch reaches a specified size.  If it returns false, the
   * automatic close should not be performed.  If it returns true,
   * the getMaxBatchSize() method should be used to obtain the
   * batch size that will trigger the automatic close.
   *
   * @return    - True if this account has been configured to
   * automatically close an open batch when the batch has reached a
   * specified size, false otherwise.
   ********************************************************************/
   public boolean isMaxBatchSizeConfigured() {
      boolean varResult = true;
      if (this.objMaxBatchSize == null) {
         varResult = false;
      }
      return varResult;
   }

   /********************************************************************
   * This method is used to get maximum batch size allowed for this
   * account's batches.  If the number of items in the batch reaches
   * this value, the cassette will automatically close the batch.
   *
   * @return    - The maximum number of batch items allowed in an open
   * batch.
   ********************************************************************/
   public int getMaxBatchSize() {

      // Assertion test:  If the maximum batch size has not been
      // configured, the value returned by this method is undefined.
      // It is considered a serious defect if this method is called
      // when the maximum batch size is not configured.  The error
      // is logged and a cassette exception is thrown.
      if (!this.isMaxBatchSizeConfigured()) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardAccount Attempting to get MaxBatchSize" +
               " but MaxBatchSize is not Configured." +
               "\n\tPRC = " + PRC_CASSETTE_ERROR +
               "\n\tSRC = " + SRC_ACCOUNT_NULL_MAX_BATCH_SIZE
            );
         }
         Vector varVector = new Vector();
         varVector.addElement(this.getMerchant());
         varVector.addElement(this.getAccountNumber());
         ErrorLog.log(
            RB_ID,
            MSG_ACCOUNT_NULL_MAX_BATCH_SIZE,
            varVector
         );
         throw new ETillCassetteException(
            ErrorLog.lookupMessage(
               RB_ID,
               MSG_ACCOUNT_NULL_MAX_BATCH_SIZE,
               varVector
            )
         );
      }
      return this.objMaxBatchSize.intValue();
   }

   /********************************************************************
   * This method is used to get an object that represents the maximum
   * batch size for this account's batches.
   *
   * @return    - Either null or a reference to an Integer object with
   * a value equal to the configured maximum batch size.
   ********************************************************************/
   public Integer getMaxBatchSizeObject () {
      return this.objMaxBatchSize;
   }


   /********************************************************************
   * This method is used to obtain a string of the object's indentifiers
   * in a form suitable for Trace.trace().  Each identifier is separated
   * with a new line character.
   *
   * @return    - A string containing the object's identifiers.
   ********************************************************************/
   public String toString() {
      return (
         "Merchant       =  " + this.getMerchant()             + "\n" +
         "Account        =  " + this.getAccountNumber()        + "\n" +
         "User           =  " + this.getUser()                 + "\n" +
         "Password       =  " + this.getObscurePassword()      + "\n" +
         "LdbCardId      =  " + this.getLdbCardId()            + "\n" +
         "Url            =  " + this.getUrl()                  + "\n" +
         "BatchCloseTime =  " + this.getBatchCloseTimeObject() + "\n" +
         "MinBatchTotal  =  " + this.getMinBatchTotalObject()  + "\n" +
         "MaxBatchSize   =  " + this.getMaxBatchSizeObject()   + "\n"
      );
   }

   /********************************************************************
   * This method constructs a Connection object used to communicate
   * with the payment engine.
   *
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   private void connect()
      throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.connect()"
         );
      }
      try {
         // Ask the DriverManager to create a Connection object.
         this.objConnection =
            DriverManager.getConnection(
               this.objUrl,
               this.objUser,
               this.objPassword
            );
      } catch (ICVException varException) {
         // The DriverManager failed to contact the RMI Server.
         // Log the error and abort.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardAccount ICVException caught." +
               "\n\tPRC = " + PRC_CASSETTE_ERROR +
               "\n\tSRC = " + SRC_ACCOUNT_ICV_FAILURE
            );
         }
         ErrorLog.logError(
            RB_ID,
            MSG_ACCOUNT_ICV_FAILURE,
            varException,
            this.getMerchant(),
            this.getAccountNumber(),
            this.getUser(),
            this.getObscurePassword(),
            this.getUrl()
         );
         // The exception may contain another exception which is
         // the original cause of the failure.  If so, we need this
         // stack trace as well.
         if (varException.getNextException() != null) {
            ErrorLog.printStackTrace(varException.getNextException());
         }
         throw new ETillAbortOperation(
            PRC_CASSETTE_ERROR,
            SRC_ACCOUNT_ICV_FAILURE
         );

      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.connect()"
         );
      }
   }

   /********************************************************************
   * This method destroys a Connection object used to communicate
   * with the payment engine.
   *
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   private void disconnect()
      throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.disconnect()"
         );
      }

      this.objConnection = null;

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.disconnect()"
         );
      }
   }

   /********************************************************************
   * This method is used by the payment object whenever it needs to
   * authorize a payment.
   *
   * @param     argCassettePayment - The cassette's payment object.
   * @return    - True if the authorization is approved, False otherwise.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public boolean authorizePayment(
      LdbCardPayment argCassettePayment
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.authorizePayment()"
         );
      }
      // Assume the authorization is declined.
      boolean varResult = false;
      // Get the framework's payment object.
      Payment varFrameworkPayment = argCassettePayment.getFrameworkPayment();
      // Get the framework's order object.
      Order   varFrameworkOrder   = varFrameworkPayment.getOrder();
      // Get the cassette's order object.
      LdbCardOrder varCassetteOrder   =
         (LdbCardOrder)varFrameworkOrder.getCassetteOrder();
      // Set the transaction type to book.
      String varType    = ICV_VAL_BOOK;
      // Set the clerk to the merchant's ID in the payment engine.
      String varClerk   = this.getLdbCardId();
      // Set the comment filed to contain the order number and
      // payment number seperated by a colon.
      String varComment =
         ICV_ORDER_PREFIX +
         varFrameworkPayment.getOrderNumber() +
         ICV_ORDER_SEPARATOR +
         varFrameworkPayment.getTransactionNumber() +
         ICV_ORDER_SUFFIX;
      // Get the credit card number from the cassette's order.
      String varPan     = varCassetteOrder.getPan();
      // Get the credit card expiration date from the cassette's order.
      String varExpiry  = varCassetteOrder.getExpiry().substring(2);
      // Get the amount as a string.
      String varAmount  = this.convertAmount(
         varFrameworkPayment.getAmount()
      );
      // If an address was recorded during the AcceptPayment request
      // processing, get it.  Otherwise use an empty string because
      // the payment engine uses an empty string to indicate that
      // the parameter is not present.
      String varAddress = varCassetteOrder.getAddress();
      if (varAddress == null) {
         varAddress = "";
      }
      // If a postal code was recorded during the AcceptPayment request
      // processing, get it.  Otherwise use an empty string because
      // the payment engine uses an empty string to indicate that
      // the parameter is not present.
      String varPostalCode = varCassetteOrder.getPostalCode();
      if (varPostalCode == null) {
         varPostalCode = "";
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceDebug(
            TRACE_ID,
            "ICVerifyTransaction.book()\n   " +
               ICV_KEY_TYPE     + " = " + varType       + "\n   " +
               ICV_KEY_CLERK    + " = " + varClerk      + "\n   " +
               ICV_KEY_COMMENT  + " = " + varComment    + "\n   " +
               ICV_KEY_ACCOUNT  + " = " +
               varCassetteOrder.getObscurePan()         + "\n   " +
               ICV_KEY_EXPIRY   + " = " + varExpiry     + "\n   " +
               ICV_KEY_AMOUNT   + " = " + varAmount     + "\n   " +
               ICV_KEY_ADDRESS  + " = " + varAddress    + "\n   " +
               ICV_KEY_ZIP_CODE + " = " + varPostalCode + "\n"
         );
      }
      // Create a statement, set each key/value pair, execute the
      // statement, and process the result.
      try {
         // Only one thread is allowed to use the connection at a
         // time.
         synchronized (this.objConnection) {
            // Ask the connection to create a new statement.
            TransactionStatement varStatement =
               this.objConnection.createTransactionStatement();
            // Set the key/value pairs.
            varStatement.set(
               ICV_KEY_TYPE,
               varType
            );
            varStatement.set(
               ICV_KEY_CLERK,
               varClerk
            );
            varStatement.set(
               ICV_KEY_COMMENT,
               varComment
            );
            varStatement.set(
               ICV_KEY_ACCOUNT,
               varPan
            );
            varStatement.set(
               ICV_KEY_EXPIRY,
               varExpiry
            );
            varStatement.set(
               ICV_KEY_AMOUNT,
               varAmount
            );
            varStatement.set(
               ICV_KEY_ADDRESS,
               varAddress
            );
            varStatement.set(
               ICV_KEY_ZIP_CODE,
               varPostalCode
            );
            // Execute the statement.
            varStatement.execute();
            // Check if the transaction was approved.
            if (varStatement.get(ICV_KEY_STATUS).equals(ICV_VAL_APPROVED)) {
               // It is approved, so record the approval code in the
               // cassette's payment object and erase any reason codes left
               // in the cassette's payment object because of a previous
               // failure.  Finally, set the result so the method returns
               // true.
               varResult = true;
               argCassettePayment.setApproval(
                  varStatement.get(ICV_KEY_APPROVAL)
               );
               argCassettePayment.setReason(null);
            } else {
               // It is declined so record the reason in the cassette's
               // payment object.  Set the result so the method returns
               // false.
               varResult = false;
               argCassettePayment.setReason(varStatement.get(ICV_KEY_REASON));
            }
            // In any case, disconnect the statement.
            varStatement.disconnect();
         }
      } catch (ICVException varException) {
         // The transaction failed.  Log the error and abort.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardAccount ICVException caught." +
               "\n\tPRC = " + PRC_CASSETTE_ERROR +
               "\n\tSRC = " + SRC_ACCOUNT_ICV_FAILURE
            );
         }
         ErrorLog.logError(
            RB_ID,
            MSG_ACCOUNT_ICV_FAILURE,
            varException,
            this.getMerchant(),
            this.getAccountNumber(),
            this.getUser(),
            this.getObscurePassword(),
            this.getUrl()
         );
         // The exception may contain another exception which is
         // the original cause of the failure.  If so, we need this
         // stack trace as well.
         if (varException.getNextException() != null) {
            ErrorLog.printStackTrace(varException.getNextException());
         }
         throw new ETillAbortOperation(
            PRC_CASSETTE_ERROR,
            SRC_ACCOUNT_ICV_FAILURE
         );
      }
      if (Trace.isAnyoneTracing()) {
         Trace.traceDebug(
            TRACE_ID,
            "ICVerifyTransaction.book()\n   " +
               ICV_KEY_APPROVAL + " = " +
               argCassettePayment.getApproval() + "\n   " +
               ICV_KEY_REASON   + " = " +
               argCassettePayment.getReason() + "\n"
         );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.authorizePayment()"
         );
      }
      return varResult;
   }

   /********************************************************************
   * This method is used by the payment object whenever it needs to
   * capture a payment in the batch.
   *
   * @param     argCassettePayment - The cassette's payment object.
   * @param     argAmount - The amount to be captured.
   * @return    - True if the authorization is approved, False otherwise.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public boolean capturePayment(
      LdbCardPayment argCassettePayment,
      int              argAmount
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.capturePayment()"
         );
      }
      // Assume the transaction is declined.
      boolean varResult = false;
      // Get the framework payment object.
      Payment varFrameworkPayment = argCassettePayment.getFrameworkPayment();
      // Get the framework order object.
      Order   varFrameworkOrder   = varFrameworkPayment.getOrder();
      // Get the cassette order object.
      LdbCardOrder varCassetteOrder   =
         (LdbCardOrder)varFrameworkOrder.getCassetteOrder();
      // Set the transaction type to Ship.
      String varType    = ICV_VAL_SHIP;
      // Set the clerk to the merchant's ID in the payment engine.
      String varClerk   = this.getLdbCardId();
      // Set the comment filed to contain the order number and
      // payment number seperated by a colon.
      String varComment =
         ICV_ORDER_PREFIX +
         varFrameworkPayment.getOrderNumber() +
         ICV_ORDER_SEPARATOR +
         varFrameworkPayment.getTransactionNumber() +
         ICV_ORDER_SUFFIX;
      // Get the credit card number from the cassette's order.
      String varPan     = varCassetteOrder.getPan();
      // Get the credit card expiration date from the cassette's order.
      String varExpiry  = varCassetteOrder.getExpiry().substring(2);
      // Get the amount as a string.
      String varAmount  = this.convertAmount(
         argAmount
      );
      // Get the approval code saved when the payment was authorized.
      String varApproval = argCassettePayment.getApproval();
      // If an address was recorded during the AcceptPayment request
      // processing, get it.  Otherwise use an empty string because
      // the payment engine uses an empty string to indicate that
      // the parameter is not present.
      String varAddress = varCassetteOrder.getAddress();
      if (varAddress == null) {
         varAddress = "";
      }
      // If a postal code was recorded during the AcceptPayment request
      // processing, get it.  Otherwise use an empty string because
      // the payment engine uses an empty string to indicate that
      // the parameter is not present.
      String varPostalCode = varCassetteOrder.getPostalCode();
      if (varPostalCode == null) {
         varPostalCode = "";
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceDebug(
            TRACE_ID,
            "ICVerifyTransaction.ship()\n   " +
               ICV_KEY_TYPE     + " = " + varType       + "\n   " +
               ICV_KEY_CLERK    + " = " + varClerk      + "\n   " +
               ICV_KEY_COMMENT  + " = " + varComment    + "\n   " +
               ICV_KEY_ACCOUNT  + " = " +
               varCassetteOrder.getObscurePan()         + "\n   " +
               ICV_KEY_EXPIRY   + " = " + varExpiry     + "\n   " +
               ICV_KEY_AMOUNT   + " = " + varAmount     + "\n   " +
               ICV_KEY_APPROVAL + " = " + varApproval   + "\n   " +
               ICV_KEY_ADDRESS  + " = " + varAddress    + "\n   " +
               ICV_KEY_ZIP_CODE + " = " + varPostalCode + "\n"
         );
      }

      // Create a statement, set each key/value pair, execute the
      // statement, and process the result.
      try {
         // Only one thread is allowed to use the connection at a
         // time.
         synchronized (this.objConnection) {
            // Ask the connection to create a new statement.
            TransactionStatement varStatement =
               this.objConnection.createTransactionStatement();
            // Set the key/value pairs.
            varStatement.set(
               ICV_KEY_TYPE,
               varType
            );
            varStatement.set(
               ICV_KEY_CLERK,
               varClerk
            );
            varStatement.set(
               ICV_KEY_COMMENT,
               varComment
            );
            varStatement.set(
               ICV_KEY_ACCOUNT,
               varPan
            );
            varStatement.set(
               ICV_KEY_EXPIRY,
               varExpiry
            );
            varStatement.set(
               ICV_KEY_AMOUNT,
               varAmount
            );
            varStatement.set(
               ICV_KEY_APPROVAL,
               varApproval
            );
            varStatement.set(
               ICV_KEY_ADDRESS,
               varAddress
            );
            varStatement.set(
               ICV_KEY_ZIP_CODE,
               varPostalCode
            );
            // Execute the statement.
            varStatement.execute();
            // Check if the transaction was approved.
            if (varStatement.get(ICV_KEY_STATUS).equals(ICV_VAL_APPROVED)) {
               // It is approved, so record the approval code in the
               // cassette's payment object and erase any reason codes left
               // in the cassette's payment object because of a previous
               // failure.  Finally, set the result so the method returns
               // true.
               varResult = true;
               argCassettePayment.setApproval(
                  varStatement.get(ICV_KEY_APPROVAL)
               );
               argCassettePayment.setReason(null);
            } else {
               // It is declined so record the reason in the cassette's
               // payment object.  Set the result so the method returns
               // false.
               varResult = false;
               argCassettePayment.setReason(varStatement.get(ICV_KEY_REASON));
            }
            // In any case, disconnect the statement.
            varStatement.disconnect();
         }
      } catch (ICVException varException) {
         // The transaction failed.  Log the error and abort.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardAccount ICVException caught." +
               "\n\tPRC = " + PRC_CASSETTE_ERROR +
               "\n\tSRC = " + SRC_ACCOUNT_ICV_FAILURE
            );
         }
         ErrorLog.logError(
            RB_ID,
            MSG_ACCOUNT_ICV_FAILURE,
            varException,
            this.getMerchant(),
            this.getAccountNumber(),
            this.getUser(),
            this.getObscurePassword(),
            this.getUrl()
         );
         // The exception may contain another exception which is
         // the original cause of the failure.  If so, we need this
         // stack trace as well.
         if (varException.getNextException() != null) {
            ErrorLog.printStackTrace(varException.getNextException());
         }
         throw new ETillAbortOperation(
            PRC_CASSETTE_ERROR,
            SRC_ACCOUNT_ICV_FAILURE
         );
      }
      if (Trace.isAnyoneTracing()) {
         Trace.traceDebug(
            TRACE_ID,
            "ICVerifyTransaction.ship()\n   " +
               ICV_KEY_APPROVAL + " = " +
               argCassettePayment.getApproval() + "\n   " +
               ICV_KEY_REASON   + " = " +
               argCassettePayment.getReason() + "\n"
         );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.capturePayment()"
         );
      }
      return varResult;
   }

   /********************************************************************
   * This method is used by the payment object whenever it needs to
   * void a payment.
   *
   * @param     argCassettePayment - The cassette's payment object.
   * @param     argAmount - The amount to be voided.
   * @return    - True if the authorization is approved, False otherwise.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public boolean voidPayment(
      LdbCardPayment   argCassettePayment,
      int              argAmount
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.voidPayment()"
         );
      }
      // Assume the transaction is declined.
      boolean varResult = false;
      // Get the framework payment object.
      Payment varFrameworkPayment = argCassettePayment.getFrameworkPayment();
      // Get the framework order object.
      Order   varFrameworkOrder   = varFrameworkPayment.getOrder();
      // Get the cassette order object.
      LdbCardOrder varCassetteOrder   =
         (LdbCardOrder)varFrameworkOrder.getCassetteOrder();
      // Set the transaction type to Void.
      String varType    = ICV_VAL_VOID;
      // Set the clerk to the merchant's ID in the payment engine.
      String varClerk   = this.getLdbCardId();
      // Set the comment filed to contain the order number and
      // payment number seperated by a colon.
      String varComment =
         ICV_ORDER_PREFIX +
         varFrameworkPayment.getOrderNumber() +
         ICV_ORDER_SEPARATOR +
         varFrameworkPayment.getTransactionNumber() +
         ICV_ORDER_SUFFIX;
      // Get the credit card number from the cassette's order.
      String varPan     = varCassetteOrder.getPan();
      // Get the credit card expiration date from the cassette's order.
      String varExpiry  = varCassetteOrder.getExpiry().substring(2);
      // Get the amount as a string.
      String varAmount  = this.convertAmount(
         argAmount
      );
      // Get the approval code saved when the payment was authorized.
      String varApproval = argCassettePayment.getApproval();
      // If an address was recorded during the AcceptPayment request
      // processing, get it.  Otherwise use an empty string because
      // the payment engine uses an empty string to indicate that
      // the parameter is not present.
      String varAddress = varCassetteOrder.getAddress();
      if (varAddress == null) {
         varAddress = "";
      }
      // If a postal code was recorded during the AcceptPayment request
      // processing, get it.  Otherwise use an empty string because
      // the payment engine uses an empty string to indicate that
      // the parameter is not present.
      String varPostalCode = varCassetteOrder.getPostalCode();
      if (varPostalCode == null) {
         varPostalCode = "";
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceDebug(
            TRACE_ID,
            "ICVerifyTransaction.void()\n   " +
               ICV_KEY_TYPE     + " = " + varType       + "\n   " +
               ICV_KEY_CLERK    + " = " + varClerk      + "\n   " +
               ICV_KEY_COMMENT  + " = " + varComment    + "\n   " +
               ICV_KEY_ACCOUNT  + " = " +
               varCassetteOrder.getObscurePan()         + "\n   " +
               ICV_KEY_EXPIRY   + " = " + varExpiry     + "\n   " +
               ICV_KEY_AMOUNT   + " = " + varAmount     + "\n   " +
               ICV_KEY_APPROVAL + " = " + varApproval   + "\n   " +
               ICV_KEY_ADDRESS  + " = " + varAddress    + "\n   " +
               ICV_KEY_ZIP_CODE + " = " + varPostalCode + "\n"
         );
      }

      // Create a statement, set each key/value pair, execute the
      // statement, and process the result.
      try {
         // Only one thread is allowed to use the connection at a
         // time.
         synchronized (this.objConnection) {
            // Ask the connection to create a new statement.
            TransactionStatement varStatement =
               this.objConnection.createTransactionStatement();
            // Set the key/value pairs.
            varStatement.set(
               ICV_KEY_TYPE,
               varType
            );
            varStatement.set(
               ICV_KEY_CLERK,
               varClerk
            );
            varStatement.set(
               ICV_KEY_COMMENT,
               varComment
            );
            varStatement.set(
               ICV_KEY_ACCOUNT,
               varPan
            );
            varStatement.set(
               ICV_KEY_EXPIRY,
               varExpiry
            );
            varStatement.set(
               ICV_KEY_AMOUNT,
               varAmount
            );
            varStatement.set(
               ICV_KEY_APPROVAL,
               varApproval
            );
            varStatement.set(
               ICV_KEY_ADDRESS,
               varAddress
            );
            varStatement.set(
               ICV_KEY_ZIP_CODE,
               varPostalCode
            );
            // Execute the statement.
            varStatement.execute();
            // Check if the transaction was approved.
            if (varStatement.get(ICV_KEY_STATUS).equals(ICV_VAL_APPROVED)) {
               // It is approved, so record the approval code in the
               // cassette's payment object and erase any reason codes left
               // in the cassette's payment object because of a previous
               // failure.  Finally, set the result so the method returns
               // true.
               varResult = true;
               argCassettePayment.setApproval(
                  varStatement.get(ICV_KEY_APPROVAL)
               );
               argCassettePayment.setReason(null);
            } else {
               // It is declined so record the reason in the cassette's
               // payment object.  Set the result so the method returns
               // false.
               varResult = false;
               argCassettePayment.setReason(varStatement.get(ICV_KEY_REASON));
            }
            // In any case, disconnect the statement.
            varStatement.disconnect();
         }
      } catch (ICVException varException) {
         // The transaction failed.  Log the error and abort.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardAccount ICVException caught." +
               "\n\tPRC = " + PRC_CASSETTE_ERROR +
               "\n\tSRC = " + SRC_ACCOUNT_ICV_FAILURE
            );
         }
         ErrorLog.logError(
            RB_ID,
            MSG_ACCOUNT_ICV_FAILURE,
            varException,
            this.getMerchant(),
            this.getAccountNumber(),
            this.getUser(),
            this.getObscurePassword(),
            this.getUrl()
         );
         // The exception may contain another exception which is
         // the original cause of the failure.  If so, we need this
         // stack trace as well.
         if (varException.getNextException() != null) {
            ErrorLog.printStackTrace(varException.getNextException());
         }
         throw new ETillAbortOperation(
            PRC_CASSETTE_ERROR,
            SRC_ACCOUNT_ICV_FAILURE
         );
      }
      if (Trace.isAnyoneTracing()) {
         Trace.traceDebug(
            TRACE_ID,
            "ICVerifyTransaction.void()\n   " +
               ICV_KEY_APPROVAL + " = " +
               argCassettePayment.getApproval() + "\n   " +
               ICV_KEY_REASON   + " = " +
               argCassettePayment.getReason() + "\n"
         );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.voidPayment()"
         );
      }
      return varResult;
   }

   /********************************************************************
   * This method is used by the credit object whenever it needs to
   * capture a credit in the batch.
   *
   * @param     argCassetteCredit - The cassette's credit object.
   * @return    - True if the authorization is approved, False otherwise.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public boolean captureCredit(
      LdbCardCredit argCassetteCredit
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.captureCredit()"
         );
      }
      // Assume the transaction is declined.
      boolean varResult = false;
      // Get the framework payment object.
      Credit  varFrameworkCredit  = argCassetteCredit.getFrameworkCredit();
      // Get the framework order object.
      Order   varFrameworkOrder   = varFrameworkCredit.getOrder();
      // Get the cassette order object.
      LdbCardOrder varCassetteOrder   =
         (LdbCardOrder)varFrameworkOrder.getCassetteOrder();
      // Set the transaction type to Credit.
      String varType    = ICV_VAL_CREDIT;
      // Set the clerk to the merchant's ID in the payment engine.
      String varClerk   = this.getLdbCardId();
      // Set the comment filed to contain the order number and
      // payment number seperated by a colon.
      String varComment =
         ICV_ORDER_PREFIX +
         varFrameworkCredit.getOrderNumber() +
         ICV_ORDER_SEPARATOR +
         varFrameworkCredit.getTransactionNumber() +
         ICV_ORDER_SUFFIX;
      // Get the credit card number from the cassette's order.
      String varPan     = varCassetteOrder.getPan();
      // Get the credit card expiration date from the cassette's order.
      String varExpiry  = varCassetteOrder.getExpiry().substring(2);
      // Get the amount as a string.
      String varAmount  = this.convertAmount(
         varFrameworkCredit.getAmount()
      );
      // If an address was recorded during the AcceptPayment request
      // processing, get it.  Otherwise use an empty string because
      // the payment engine uses an empty string to indicate that
      // the parameter is not present.
      String varAddress = varCassetteOrder.getAddress();
      if (varAddress == null) {
         varAddress = "";
      }
      // If a postal code was recorded during the AcceptPayment request
      // processing, get it.  Otherwise use an empty string because
      // the payment engine uses an empty string to indicate that
      // the parameter is not present.
      String varPostalCode = varCassetteOrder.getPostalCode();
      if (varPostalCode == null) {
         varPostalCode = "";
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceDebug(
            TRACE_ID,
            "ICVerifyTransaction.credit()\n   " +
               ICV_KEY_TYPE     + " = " + varType       + "\n   " +
               ICV_KEY_CLERK    + " = " + varClerk      + "\n   " +
               ICV_KEY_COMMENT  + " = " + varComment    + "\n   " +
               ICV_KEY_ACCOUNT  + " = " +
               varCassetteOrder.getObscurePan()         + "\n   " +
               ICV_KEY_EXPIRY   + " = " + varExpiry     + "\n   " +
               ICV_KEY_AMOUNT   + " = " + varAmount     + "\n   " +
               ICV_KEY_ADDRESS  + " = " + varAddress    + "\n   " +
               ICV_KEY_ZIP_CODE + " = " + varPostalCode + "\n"
         );
      }

      // Create a statement, set each key/value pair, execute the
      // statement, and process the result.
      try {
         // Only one thread is allowed to use the connection at a
         // time.
         synchronized (this.objConnection) {
            // Ask the connection to create a new statement.
            TransactionStatement varStatement =
               this.objConnection.createTransactionStatement();
            // Set the key/value pairs.
            varStatement.set(
               ICV_KEY_TYPE,
               varType
            );
            varStatement.set(
               ICV_KEY_CLERK,
               varClerk
            );
            varStatement.set(
               ICV_KEY_COMMENT,
               varComment
            );
            varStatement.set(
               ICV_KEY_ACCOUNT,
               varPan
            );
            varStatement.set(
               ICV_KEY_EXPIRY,
               varExpiry
            );
            varStatement.set(
               ICV_KEY_AMOUNT,
               varAmount
            );
            varStatement.set(
               ICV_KEY_ADDRESS,
               varAddress
            );
            varStatement.set(
               ICV_KEY_ZIP_CODE,
               varPostalCode
            );
            // Execute the statement.
            varStatement.execute();
            // Check if the transaction was approved.
            if (varStatement.get(ICV_KEY_STATUS).equals(ICV_VAL_APPROVED)) {
               // It is approved, so record the approval code in the
               // cassette's credit object and erase any reason codes left
               // in the cassette's credit object because of a previous
               // failure.  Finally, set the result so the method returns
               // true.
               varResult = true;
               argCassetteCredit.setApproval(
                  varStatement.get(ICV_KEY_APPROVAL)
               );
               argCassetteCredit.setReason(null);
            } else {
               // It is declined so record the reason in the cassette's
               // credit object.  Set the result so the method returns
               // false.
               varResult = false;
               argCassetteCredit.setReason(varStatement.get(ICV_KEY_REASON));
            }
            // In any case, disconnect the statement.
            varStatement.disconnect();
         }
      } catch (ICVException varException) {
         // The transaction failed.  Log the error and abort.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardAccount ICVException caught." +
               "\n\tPRC = " + PRC_CASSETTE_ERROR +
               "\n\tSRC = " + SRC_ACCOUNT_ICV_FAILURE
            );
         }
         ErrorLog.logError(
            RB_ID,
            MSG_ACCOUNT_ICV_FAILURE,
            varException,
            this.getMerchant(),
            this.getAccountNumber(),
            this.getUser(),
            this.getObscurePassword(),
            this.getUrl()
         );
         // The exception may contain another exception which is
         // the original cause of the failure.  If so, we need this
         // stack trace as well.
         if (varException.getNextException() != null) {
            ErrorLog.printStackTrace(varException.getNextException());
         }
         throw new ETillAbortOperation(
            PRC_CASSETTE_ERROR,
            SRC_ACCOUNT_ICV_FAILURE
         );
      }
      if (Trace.isAnyoneTracing()) {
         Trace.traceDebug(
            TRACE_ID,
            "ICVerifyTransaction.credit()\n   " +
               ICV_KEY_APPROVAL + " = " +
               argCassetteCredit.getApproval() + "\n   " +
               ICV_KEY_REASON   + " = " +
               argCassetteCredit.getReason() + "\n"
         );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.captureCredit()"
         );
      }
      return varResult;
   }

   /********************************************************************
   * This method is used by the credit object whenever it needs to
   * void a credit.
   *
   * @param     argCassetteCredit - The cassette's credit object.
   * @return    - True if the authorization is approved, False otherwise.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public boolean voidCredit(
      LdbCardCredit argCassetteCredit
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.voidCredit()"
         );
      }
      // Assume the transaction is declined.
      boolean varResult = false;
      // Get the framework payment object.
      Credit  varFrameworkCredit  = argCassetteCredit.getFrameworkCredit();
      // Get the framework order object.
      Order   varFrameworkOrder   = varFrameworkCredit.getOrder();
      // Get the cassette order object.
      LdbCardOrder varCassetteOrder   =
         (LdbCardOrder)varFrameworkOrder.getCassetteOrder();
      // Set the transaction type to VoidCredit.
      String varType    = ICV_VAL_VOID_C;
      // Set the clerk to the merchant's ID in the payment engine.
      String varClerk   = this.getLdbCardId();
      // Set the comment filed to contain the order number and
      // payment number seperated by a colon.
      String varComment =
         ICV_ORDER_PREFIX +
         varFrameworkCredit.getOrderNumber() +
         ICV_ORDER_SEPARATOR +
         varFrameworkCredit.getTransactionNumber() +
         ICV_ORDER_SUFFIX;
      // Get the credit card number from the cassette's order.
      String varPan     = varCassetteOrder.getPan();
      // Get the credit card expiration date from the cassette's order.
      String varExpiry  = varCassetteOrder.getExpiry().substring(2);
      // Get the amount as a string.
      String varAmount  = this.convertAmount(
         varFrameworkCredit.getAmount()
      );
      // Get the approval code saved when the credit was authorized.
      String varApproval = argCassetteCredit.getApproval();
      // If an address was recorded during the AcceptPayment request
      // processing, get it.  Otherwise use an empty string because
      // the payment engine uses an empty string to indicate that
      // the parameter is not present.
      String varAddress = varCassetteOrder.getAddress();
      if (varAddress == null) {
         varAddress = "";
      }
      // If a postal code was recorded during the AcceptPayment request
      // processing, get it.  Otherwise use an empty string because
      // the payment engine uses an empty string to indicate that
      // the parameter is not present.
      String varPostalCode = varCassetteOrder.getPostalCode();
      if (varPostalCode == null) {
         varPostalCode = "";
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceDebug(
            TRACE_ID,
            "ICVerifyTransaction.voidcredit()\n   " +
               ICV_KEY_TYPE     + " = " + varType       + "\n   " +
               ICV_KEY_CLERK    + " = " + varClerk      + "\n   " +
               ICV_KEY_COMMENT  + " = " + varComment    + "\n   " +
               ICV_KEY_ACCOUNT  + " = " +
               varCassetteOrder.getObscurePan()         + "\n   " +
               ICV_KEY_EXPIRY   + " = " + varExpiry     + "\n   " +
               ICV_KEY_AMOUNT   + " = " + varAmount     + "\n   " +
               ICV_KEY_APPROVAL + " = " + varApproval   + "\n   " +
               ICV_KEY_ADDRESS  + " = " + varAddress    + "\n   " +
               ICV_KEY_ZIP_CODE + " = " + varPostalCode + "\n"
         );
      }

      // Create a statement, set each key/value pair, execute the
      // statement, and process the result.
      try {
         // Only one thread is allowed to use the connection at a
         // time.
         synchronized (this.objConnection) {
            // Ask the connection to create a new statement.
            TransactionStatement varStatement =
               this.objConnection.createTransactionStatement();
            // Set the key/value pairs.
            varStatement.set(
               ICV_KEY_TYPE,
               varType
            );
            varStatement.set(
               ICV_KEY_CLERK,
               varClerk
            );
            varStatement.set(
               ICV_KEY_COMMENT,
               varComment
            );
            varStatement.set(
               ICV_KEY_ACCOUNT,
               varPan
            );
            varStatement.set(
               ICV_KEY_EXPIRY,
               varExpiry
            );
            varStatement.set(
               ICV_KEY_AMOUNT,
               varAmount
            );
            varStatement.set(
               ICV_KEY_APPROVAL,
               varApproval
            );
            varStatement.set(
               ICV_KEY_ADDRESS,
               varAddress
            );
            varStatement.set(
               ICV_KEY_ZIP_CODE,
               varPostalCode
            );
            // Execute the statement.
            varStatement.execute();
            // Check if the transaction was approved.
            if (varStatement.get(ICV_KEY_STATUS).equals(ICV_VAL_APPROVED)) {
               // It is approved, so record the approval code in the
               // cassette's credit object and erase any reason codes left
               // in the cassette's credit object because of a previous
               // failure.  Finally, set the result so the method returns
               // true.
               varResult = true;
               argCassetteCredit.setApproval(
                  varStatement.get(ICV_KEY_APPROVAL)
               );
               argCassetteCredit.setReason(null);
            } else {
               // It is declined so record the reason in the cassette's
               // credit object.  Set the result so the method returns
               // false.
               varResult = false;
               argCassetteCredit.setReason(varStatement.get(ICV_KEY_REASON));
            }
            // In any case, disconnect the statement.
            varStatement.disconnect();
         }
      } catch (ICVException varException) {
         // The transaction failed.  Log the error and abort.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardAccount ICVException caught." +
               "\n\tPRC = " + PRC_CASSETTE_ERROR +
               "\n\tSRC = " + SRC_ACCOUNT_ICV_FAILURE
            );
         }
         ErrorLog.logError(
            RB_ID,
            MSG_ACCOUNT_ICV_FAILURE,
            varException,
            this.getMerchant(),
            this.getAccountNumber(),
            this.getUser(),
            this.getObscurePassword(),
            this.getUrl()
         );
         // The exception may contain another exception which is
         // the original cause of the failure.  If so, we need this
         // stack trace as well.
         if (varException.getNextException() != null) {
            ErrorLog.printStackTrace(varException.getNextException());
         }
         throw new ETillAbortOperation(
            PRC_CASSETTE_ERROR,
            SRC_ACCOUNT_ICV_FAILURE
         );
      }
      if (Trace.isAnyoneTracing()) {
         Trace.traceDebug(
            TRACE_ID,
            "ICVerifyTransaction.voidcredit()\n   " +
               ICV_KEY_APPROVAL + " = " +
               argCassetteCredit.getApproval() + "\n   " +
               ICV_KEY_REASON   + " = " +
               argCassetteCredit.getReason() + "\n"
         );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.voidCredit()"
         );
      }
      return varResult;
   }

   /********************************************************************
   * This method is used by the batch object whenever it needs to
   * settle a batch.
   *
   * @param     argBatch - The cassette's batch object.
   * @return    - True if the settlement succeeds, False otherwise.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public boolean settleBatch(
      LdbCardBatch  argCassetteBatch
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardAccount.settleBatch()"
         );
      }
      // Assume the transaction is declined.
      boolean varResult = false;
      // Set the transaction type to Settle;
      String varType    = ICV_VAL_SETTLE;
      // Set the clerk to the merchant's ID in the payment engine.
      String varClerk   = this.getLdbCardId();

      if (Trace.isAnyoneTracing()) {
         Trace.traceDebug(
            TRACE_ID,
            "ICVerifyTransaction.settleBatch()\n   " +
               ICV_KEY_TYPE     + " = " + varType       + "\n   " +
               ICV_KEY_CLERK    + " = " + varClerk      + "\n"
         );
      }

      // Create a statement, set each key/value pair, execute the
      // statement, and process the result.
      try {
         // Only one thread is allowed to use the connection at a
         // time.
         synchronized (this.objConnection) {
            // Ask the connection to create a new statement.
            TransactionStatement varStatement =
               this.objConnection.createTransactionStatement();
            // Set the key/value pairs.
            varStatement.set(
               ICV_KEY_TYPE,
               varType
            );
            varStatement.set(
               ICV_KEY_CLERK,
               varClerk
            );
            // Execute the statement.
            varStatement.execute();
            // Check if the transaction was approved.
            if (varStatement.get(ICV_KEY_STATUS).equals(ICV_VAL_APPROVED)) {
               // It is approved, so record the approval code in the
               // cassette's batch object and erase any reason codes left
               // in the cassette's batch object because of a previous
               // failure.  Finally, set the result so the method returns
               // true.
               varResult = true;
               argCassetteBatch.setApproval(
                  varStatement.get(ICV_KEY_APPROVAL)
               );
               argCassetteBatch.setReason(null);
            } else {
               // It is declined so record the reason in the cassette's
               // credit object.  Set the result so the method returns
               // false.
               varResult = false;
               argCassetteBatch.setReason(varStatement.get(ICV_KEY_REASON));
            }
            // In any case, disconnect the statement.
            varStatement.disconnect();
         }
      } catch (ICVException varException) {
         // The transaction failed.  Log the error and abort.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardAccount ICVException caught." +
               "\n\tPRC = " + PRC_CASSETTE_ERROR +
               "\n\tSRC = " + SRC_ACCOUNT_ICV_FAILURE
            );
         }
         ErrorLog.logError(
            RB_ID,
            MSG_ACCOUNT_ICV_FAILURE,
            varException,
            this.getMerchant(),
            this.getAccountNumber(),
            this.getUser(),
            this.getObscurePassword(),
            this.getUrl()
         );
         // The exception may contain another exception which is
         // the original cause of the failure.  If so, we need this
         // stack trace as well.
         if (varException.getNextException() != null) {
            ErrorLog.printStackTrace(varException.getNextException());
         }
         throw new ETillAbortOperation(
            PRC_CASSETTE_ERROR,
            SRC_ACCOUNT_ICV_FAILURE
         );
      }
      if (Trace.isAnyoneTracing()) {
         Trace.traceDebug(
            TRACE_ID,
            "ICVerifyTransaction.settle()\n   " +
               ICV_KEY_APPROVAL + " = " +
               argCassetteBatch.getApproval() + "\n   " +
               ICV_KEY_REASON   + " = " +
               argCassetteBatch.getReason() + "\n"
         );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardAccount.settleBatch()"
         );
      }
      return varResult;
   }

   /********************************************************************
   * This method is used to convert a framework amount into a
   * string that is acceptable to the payment engine.  The account's
   * data formatter has been initialized to understand the currency
   * symbol and the decimal separator required for the currency used.
   *
   * @param     argAmount - The amount to be converted.
   * @return    - The amount converted into a string acceptable by the
   * payment engine.
   ********************************************************************/
   private String convertAmount(
      int argAmount
   ) {
      return (
         this.objFormatter.format(
            (double)argAmount * Math.pow((double)10, this.objDoubleExp10)
         )
      );
   }
}
