/*****************************************************************************
*
*  QueryTest
*
*****************************************************************************/

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringReader;
import java.util.Enumeration;
import java.util.Hashtable;

import com.ibm.etill.framework.clientapi.PaymentCommandConstants;
import com.ibm.etill.framework.clientapi.PaymentServerClient;
import com.ibm.etill.framework.clientapi.PaymentServerResponse;
import com.ibm.etill.framework.payapi.FrameworkReturnCodes;

import com.ibm.etill.xml.parser.Parser;
import com.ibm.etill.xml.parser.TXDocument;
import com.ibm.etill.org.w3c.dom.Document;

import HTMLPage;

/***********************************************************************
* The QueryTest class is a simple servlet that can be used to
* generate Payment Manage orders using the AcceptPayment command.
* <p>
* The HTTP POST used to trigger this servlet should contain the
* following parameter keywords.
* <ul>
* <li> PAYMENTTYPE
* <li> MERCHANTNUMBER
* <li> AMOUNT
* <li> CURRENCY.
* </ul>
* <p>
* If the parameter keywords EXPIRYMONTH and EXPIRYYEAR are provided, the
* values will be combined and set for the $EXPIRY command keyword.
* <p>
* This servlet will use the date/time to generate an order number.
* <p>
* For any parameter keyword that begins with a $, the keyword and
* its value will be passed to the Payment Manager.
***********************************************************************/
public class QueryTest
   extends HttpServlet
   implements PaymentCommandConstants, FrameworkReturnCodes {

   /********************************************************************
   * This servlet attribute contains the name of the Payment Manager
   * host.  Normally, it would be initialized from configuration
   * (e.g. a servlet parameter).
   ********************************************************************/
   String objPaymentManagerHost = "localhost";
   /********************************************************************
   * This servlet attribute contains the port where the Payment Manager
   * listens.  Normally, it would be initialized from configuration
   * (e.g. a servlet parameter).
   ********************************************************************/
   int    objPaymentManagerPort = 80;
   /********************************************************************
   * This servlet attribute contains the path to XML Document Type
   * Definitions.  Normally, it would be initialized from configuration
   * (e.g. a servlet parameter).
   ********************************************************************/
   String objDtdPath = null;
   /********************************************************************
   * This servlet attribute contains the user ID to be used when
   * connecting.  Normally, it would be initialized from configuration
   * (e.g. a servlet parameter).
   ********************************************************************/
   String objUserId = "admin";
   /********************************************************************
   * This servlet attribute contains the password used to authenticate
   * the user.  Normally, it would be initialized from configuration
   * (e.g. a servlet parameter).
   ********************************************************************/
   String objPassword = "admin";
   /********************************************************************
   * This servlet attribute contains the client object constructed
   * during initializaiton.
   ********************************************************************/
   PaymentServerClient objPaymentManager = null;

   /********************************************************************
   * This method is called by the servlet manager after the servlet
   * class is loaded and the servlet object is created but before
   * any post requests are passed to it.
   * <p>
   * For this servlet, it simply initializes the object used to
   * communicate with the Payment Manager.
   *
   * @param     argConfiguration - The servlet's configuration
   * environment.
   * @exception ServletException - Any error results in this
   * exception.
   ********************************************************************/
   public void init(
                   ServletConfig argConfiguration
                   )  throws ServletException {
      super.init(argConfiguration);

      // Construct client object.
      objPaymentManager = new PaymentServerClient(
         this.objDtdPath,
         this.objPaymentManagerHost,
         this.objPaymentManagerPort
      );
   }


   /********************************************************************
   * This method is called by the servlet manager every time an HTTP
   * POST request is sent to this servlet.
   * <p>
   * For this servlet, it uses the CGI parameters passed on the POST
   * request to build an AcceptPayment request for the Payment Manager.
   * Then it builds a dynamic HTML page as a response.
   *
   * @param     argRequest - The servlet request that triggered this
   * call.
   * @param     argResponse - The servlet response used to reply.
   * @exception ServletException - processing errors throw this
   * exception.
   * @exception IOException - I/O errors throw this exception.
   ********************************************************************/
   public void doPost(
      HttpServletRequest  argRequest,
      HttpServletResponse argResponse
   ) throws ServletException, IOException{


      argResponse.setContentType("text/html");
      PrintWriter varClient = argResponse.getWriter();

      try {

         // Construct a table to contain the command's parameters.
         Hashtable varParameters = new Hashtable();

         // Create a string reference for convenience.
         String varValue = null;
         String varCommand = null;

         // For each key provided by the CGI form, add the key and its
         // value to the parameter list.  There is a special case when
         // the CGI parameter keyword is COMMAND.  This value is not
         // placed in the parameter list.  It is saved as the command
         // name instead.
         Enumeration varKeys = argRequest.getParameterNames();
         while (varKeys.hasMoreElements()) {
            String varKey = (String) varKeys.nextElement();
            varValue = this.getServletParameter(
               argRequest,
               varKey
            );
            if (varKey.equals("COMMAND")) {
               varCommand = varValue;
            } else if (varValue != null) {
               varParameters.put(
                  varKey,
                  varValue
               );
            }
         }

         PaymentServerResponse varResponse = null;
         String varHttpCommand = null;
         StringReader varStream = null;

         // Must synchronize on the Payment Manager object because the
         // getLastCommand() and getLastXML() methods are not thread
         // safe.
         synchronized (this.objPaymentManager) {

            // Issue the command.
            varResponse =
               this.objPaymentManager.issueCommand(
                  varCommand,
                  varParameters,
                  this.objUserId,
                  this.objPassword
               );

            // Get the resulting HTTP Command.

            varHttpCommand = this.objPaymentManager.getLastRequest();

            // Get the XML Document.
            varStream =
               new StringReader(
                  this.objPaymentManager.getLastXML()
               );
         }

         // Construct an XML document and parse it into a document.
         Parser varParser = new Parser("XML Response Parser");
         Document varDocument = varParser.readStream(varStream);

         // Display the result.
         this.displayResultToClient(
            varResponse.getPrimaryRC(),
            varResponse.getSecondaryRC(),
            varResponse.getObjectCount(),
            varParser.getNumberOfErrors(),
            varParser.getNumberOfWarnings(),
            varHttpCommand,
            (TXDocument)varDocument,
            varClient
         );


      } catch (Exception varException) {
         this.displayErrorToClient(
            varException,
            varClient
         );
      }

   }

   /********************************************************************
   * The servlet uses this method to extract a single keyword value from
   * the servlet request.
   *
   * @param     argRequest - The servlet request.
   * @param     argKey - The parameter key whose value is to be
   * returned.
   * @return    - The first value found for the given keyword or null
   * if the value is an empty string.
   ********************************************************************/
   private String getServletParameter(
      HttpServletRequest  argRequest,
      String              argKey
   ) {
      String[] varValues = argRequest.getParameterValues(argKey);
      String varResult = varValues[0];
      if (varResult.length() == 0) {
         varResult = null;
      }
      return varResult;
   }


   /*******************************************************************
   * This method is used to display an error HTML page on the user's
   * browser.
   *
   * @param     argException - The exception generated when the error
   * was detected.
   * @param     argClient - A servlet print stream where the HTML
   * page is to be written.
   *******************************************************************/
   private void displayErrorToClient(
      Exception   argException,
      PrintWriter argClient
   ) {
      String varMessage = argException.getMessage();

      if (varMessage == null) {
         varMessage = "Internal Error - Programming Defect";
      }

      HTMLPage varPage = new HTMLPage(argClient);

      varPage.StartPage();
      varPage.StartHead();
      varPage.Title("Sample Store" + " Error Message");
      varPage.EndHead();
      varPage.StartBody("#ffffff");

      varPage.HeadingThree("Sample Store" + " Error Message");

      varPage.StartTable(0);
      varPage.StartTableRow("left", "top");
      varPage.TableHeader("ERROR: ");
      varPage.TableData(varMessage);
      varPage.EndTableRow();
      varPage.EndTable();
      varPage.PreFormatted(argException);
      varPage.Paragraph("Press the 'back button' on your browser and correct " +
                  "the problem");

      varPage.EndBody();
      varPage.EndPage();
   }

   /*******************************************************************
   * This method is used to display the result of a command to the
   * tester's browser.
   *
   * @param     argPrimaryRc - The primary return code returned by
   * the Payment Manager.
   * @param     argSecondaryRc - The secondary return code returned
   * by the Payment Manager.
   * @param     argClient - A servlet print stream where the HTML
   * page is to be written.
   ********************************************************************/
   private void displayResultToClient(
      short       argPrimaryRc,
      short       argSecondaryRc,
      int         argObjectCount,
      int         argErrors,
      int         argWarnings,
      String      argHttpCommand,
      TXDocument  argDocument,
      PrintWriter argClient
   ) throws IOException {

      String varHeader  = "Error:";
      String varMessage = "The requested query has failed.";
      if (argPrimaryRc == PRC_OPERATION_SUCCESS) {
         if (argSecondaryRc == RC_NONE) {
            varHeader  = "Successful:";
            varMessage = "The requested query has been completed.";
         } else {
            varHeader  = "Warning:";
            varMessage = "The requested query has been completed " +
               "but the secondary return code indicates a warning.";
         }
      }

      HTMLPage varPage = new HTMLPage(argClient);

      varPage.StartPage();
      varPage.StartHead();
      varPage.Title("Query Test" + " Result Message");
      varPage.EndHead();
      varPage.StartBody("#ffffff");

      varPage.HeadingThree("Query Test" + " Result Message");

      varPage.StartTable(0);
      varPage.StartTableRow("left", "top");
      varPage.TableHeader(varHeader);
      varPage.TableData(varMessage);
      varPage.EndTableRow();
      varPage.StartTableRow("left", "top");
      varPage.TableHeader("Primary Return Code ==>");
      varPage.TableData(Short.toString(argPrimaryRc), "right", "top");
      varPage.EndTableRow();
      varPage.StartTableRow("left", "top");
      varPage.TableHeader("Secondary Return Code ==>");
      varPage.TableData(Short.toString(argSecondaryRc), "right", "top");
      varPage.EndTableRow();
      varPage.StartTableRow("left", "top");
      varPage.TableHeader("Object count ==>");
      varPage.TableData(Integer.toString(argObjectCount), "right", "top");
      varPage.EndTableRow();
      varPage.StartTableRow("left", "top");
      varPage.TableHeader("Number of XML Parser Errors ==>");
      varPage.TableData(Integer.toString(argErrors), "right", "top");
      varPage.EndTableRow();
      varPage.StartTableRow("left", "top");
      varPage.TableHeader("Number of XML Parser Warnings ==>");
      varPage.TableData(Integer.toString(argWarnings), "right", "top");
      varPage.EndTableRow();
      varPage.EndTable();

      varPage.HeadingThree("Query Test" + " Command Issued");

      varPage.PreFormatted(argHttpCommand);

      varPage.HeadingThree("Query Test" + " XML Response");

      argClient.println("<XMP>");
      argDocument.print(argClient);
      argClient.println("\n</XMP>");
      varPage.Paragraph("Press the 'back button' on your browser " +
                  "to repeat the operation");

      varPage.EndBody();
      varPage.EndPage();
   }
}
